// Copyright (c) 1999-2001  David Muse
// See the file COPYING for more information

#include <config.h>
#include <defaults.h>
#include <datatypes.h>
#include <defines.h>
#include <sqlrelay/sqlrclient.h>

#include <iomanip.h>

#include <math.h>
#include <time.h>
#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#ifdef HAVE_STRINGS_H
	#include <strings.h>
#endif
#include <sys/types.h>
#include <sys/stat.h>
extern "C" {
	#undef __cplusplus
	#include <fcntl.h>
	#define __cplusplus
}
#ifdef HAVE_UNISTD_H
	#include <unistd.h>
#endif

sqlrconnection::sqlrconnection(char *server, int port, char *socket,
					char *user, char *password, 
					int retrytime, int tries) 
: clientsocket() {

	// initialize...

	// retry reads if they get interrupted by signals
	retryInterruptedReads();

	// connection
	this->server=server;
	listenerinetport=(unsigned short)port;
	listenerunixport=socket;
	this->retrytime=retrytime;
	this->tries=tries;

	// authentication
	this->user=user;
	this->password=password;
	userlen=strlen(user);
	passwordlen=strlen(password);
	reconnect=0;

	// database id
	id=NULL;

	// session state
	connected=0;
	clearSessionFlags();

	// debug print function
	printfunction=NULL;

	// debug off
	debug=0;
	webdebug=-1;

	// copy references, delete cursors flags
	copyrefs=0;

	// error string
	error=NULL;

	// cursor list
	firstcursor=NULL;
	lastcursor=NULL;
}

sqlrconnection::~sqlrconnection() {

	// unless it was already ended or suspended, end the session
	if (!endsessionsent && !suspendsessionsent) {
		endSession();
	}

	// deallocate id
	if (id) {
		delete[] id;
	}

	// deallocate copied references
	if (copyrefs) {
		if (server) {
			delete[] server;
		}
		if (listenerunixport) {
			delete[] listenerunixport;
		}
		if (user) {
			delete[] user;
		}
		if (password) {
			delete[] password;
		}
	}

	// detach all cursors attached to this client
	sqlrcursor	*currentcursor=firstcursor;
	while (currentcursor) {
		firstcursor=currentcursor;
		currentcursor=currentcursor->next;
		firstcursor->sqlrc=NULL;
	}

	if (debug) {
		debugPreStart();
		debugPrint("Deallocated connection\n");
		debugPreEnd();
	}
}

int	sqlrconnection::endSession() {

	if (debug) {
		debugPreStart();
		debugPrint("Ending Session\n");
		debugPreEnd();
	}

	// abort each cursor's result set
	sqlrcursor	*currentcursor=firstcursor;
	while (currentcursor) {
		if (!currentcursor->endofresultset) {
			currentcursor->abortResultSet();
		}
		currentcursor=currentcursor->next;
	}

	// write a ~ to the connection
	int	retval=1;
	if (connected) {
		retval=0;
		writeData((unsigned short)END_SESSION);
		endsessionsent=1;
		retval=1;
		closeConnection();
	}
	return retval;
}

int	sqlrconnection::suspendSession() {

	if (!connected && !openSession()) {
		return 0;
	}

	if (debug) {
		debugPreStart();
		debugPrint("Suspending Session\n");
		debugPreEnd();
	}

	// suspend the session
	int	retval=0;
	writeData((unsigned short)SUSPEND_SESSION);
	suspendsessionsent=1;
	retval=1;

	// If the server is passing around file descriptors to handoff clients
	// from listener to connection, then it will have to open a socket and
	// port to enable suspend/resume.   It will pass that socket/port to
	// us here.
	if (!reconnect) {
		retval=getNewPort();
	}

	closeConnection();

	return retval;
}

int	sqlrconnection::getConnectionPort() {

	if (!suspendsessionsent && !connected && !openSession()) {
		return 0;
	}

	if (debug) {
		debugPreStart();
		debugPrint("Getting connection port: ");
		debugPrint((long)connectioninetport);
		debugPrint("\n");
		debugPreEnd();
	}

	return (int)connectioninetport;
}

char	*sqlrconnection::getConnectionSocket() {

	if (!suspendsessionsent && !connected && !openSession()) {
		return NULL;
	}

	if (debug) {
		debugPreStart();
		debugPrint("Getting connection socket: ");
		if (connectionunixport) {
			debugPrint(connectionunixport);
		}
		debugPrint("\n");
		debugPreEnd();
	}

	if (connectionunixport) {
		return connectionunixport;
	}
	return NULL;
}

int	sqlrconnection::authenticateWithListener() {
	if (debug) {
		debugPreStart();
		debugPrint("Authenticating with listener : ");
	}
	return genericAuthentication();
}

int	sqlrconnection::genericAuthentication() {

	if (debug) {
		debugPrint(user);
		debugPrint(":");
		debugPrint(password);
		debugPrint("\n");
		debugPreEnd();
	}

	writeData((unsigned long)userlen);
	writeData(user,userlen);

	writeData((unsigned long)passwordlen);
	writeData(password,passwordlen);

	// check whether authentication was successful or not
	unsigned short	authsuccess;
	if (readData(&authsuccess)!=sizeof(unsigned short)) {
		setError("Failed to authenticate.\n A network error may have ocurred.");
		return -1;
	}
	if (authsuccess==ERROR) {
		
		// clear all result sets
		sqlrcursor	*currentcursor=firstcursor;
		while (currentcursor) {
			currentcursor->clearResultSet();
			currentcursor=currentcursor->next;
		}

		setError("Authentication Error.");
		return 0;
	}
	if (debug) {
		debugPreStart();
		debugPrint("No authentication error.\n");
		debugPreEnd();
	}
	return 1;
}

void	sqlrconnection::getReconnect() {

	unsigned short	recon;
	if (readData(&recon)!=sizeof(unsigned short)) {
		reconnect=-1;
		return;
	}

	if (recon==DONT_RECONNECT) {
		if (debug) {
			debugPreStart();
			debugPrint("Must Not Reconnect.\n");
			debugPreEnd();
		}
		reconnect=0;
		return;
	}
	if (debug) {
		debugPreStart();
		debugPrint("Must Reconnect.\n");
		debugPreEnd();
	}
	reconnect=1;
}

int	sqlrconnection::getNewPort() {

	// get the size of the unix port string
	unsigned short	size;
	if (readData(&size)!=sizeof(unsigned short)) {
		return -1;
	}

	if (size<=MAXPATHLEN) {

		// get the unix port string
		if (size && readData(connectionunixportbuffer,size)!=size) {
			return -1;
		}
		connectionunixportbuffer[size]=(char)NULL;
		connectionunixport=connectionunixportbuffer;

	} else {

		// if size is too big, return an error
		strstream	*errstr=new strstream();
		*errstr << "The pathname of the unix port was too long: ";
		*errstr << size << " bytes.  The maximum size is ";
		*errstr << MAXPATHLEN << " bytes." << ENDS;
		setError(errstr->str());
		return 0;

	}

	// get the inet port
	if (readData((unsigned short *)&connectioninetport)!=
					sizeof(unsigned short)) {
		return -1;
	}

	return 1;
}

int	sqlrconnection::ping() {

	if (!connected && !openSession()) {
		return 0;
	}

	if (debug) {
		debugPreStart();
		debugPrint("Pinging...");
		debugPrint("\n");
		debugPreEnd();
	}

	writeData((unsigned short)PING);

	// get the ping result
	unsigned short	result;
	if (readData(&result)!=sizeof(unsigned short)) {
		setError("Failed to ping.\n A network error may have ocurred.");
		result=0;
	}
	return (int)result;
}


char	*sqlrconnection::identify() {

	if (!connected && !openSession()) {
		return 0;
	}

	if (debug) {
		debugPreStart();
		debugPrint("Identifying...");
		debugPrint("\n");
		debugPreEnd();
	}

	writeData((unsigned short)IDENTIFY);

	// get the id
	unsigned short	size;
	if (readData(&size)==sizeof(unsigned short)) {
		id=new char[size+1];
		if (readData(id,size)!=size) {
			setError("Failed to identify.\n A network error may have ocurred.");
			delete[] id;
			return NULL;
		}
		id[size]=(char)NULL;

		if (debug) {
			debugPreStart();
			debugPrint(id);
			debugPrint("\n");
			debugPreEnd();
		}
	} else {
		setError("Failed to identify.\n A network error may have ocurred.");
		return NULL;
	}
	return id;
}

int	sqlrconnection::autoCommitOn() {
	return autoCommit(1);
}

int	sqlrconnection::autoCommitOff() {
	return autoCommit(0);
}

int	sqlrconnection::autoCommit(unsigned short on) {

	if (!connected && !openSession()) {
		return 0;
	}

	if (debug) {
		debugPreStart();
		debugPrint("Setting AutoCommit");
		if (on) {
			debugPrint("on");
		} else {
			debugPrint("off");
		}
		debugPrint("...\n");
		debugPreEnd();
	}

	writeData((unsigned short)AUTOCOMMIT);
	writeData(on);

	unsigned short	response;
	if (readData(&response)!=sizeof(unsigned short)) {
		setError("Failed to get autocommit status.\n A network error may have ocurred.");
		return -1;
	}
	return (int)response;
}

int	sqlrconnection::commit() {

	if (!connected && !openSession()) {
		return 0;
	}

	if (debug) {
		debugPreStart();
		debugPrint("Committing...");
		debugPrint("\n");
		debugPreEnd();
	}

	writeData((unsigned short)COMMIT);

	unsigned short	response;
	if (readData(&response)!=sizeof(unsigned short)) {
		setError("Failed to get commit status.\n A network error may have ocurred.");
		return -1;
	}
	return (int)response;
}

int	sqlrconnection::rollback() {

	if (!connected && !openSession()) {
		return 0;
	}

	if (debug) {
		debugPreStart();
		debugPrint("Rolling Back...");
		debugPrint("\n");
		debugPreEnd();
	}

	writeData((unsigned short)ROLLBACK);

	unsigned short	response;
	if (readData(&response)!=sizeof(unsigned short)) {
		setError("Failed to get rollback status.\n A network error may have ocurred.");
		return -1;
	}
	return (int)response;
}

void	sqlrconnection::debugOn() {
	debug=1;
}

void	sqlrconnection::debugOff() {
	debug=0;
}

int	sqlrconnection::getDebug() {
	return debug;
}

void	sqlrconnection::clearError() {
	if (error) {
		delete[] error;
		error=NULL;
	}
}

void	sqlrconnection::setError(char *err) {

	if (debug) {
		debugPreStart();
		debugPrint("Setting Error\n");
		debugPreEnd();
	}

	error=new char[strlen(err)+1];
	strcpy(error,err);

	if (debug) {
		debugPreStart();
		debugPrint(error);
		debugPrint("\n");
		debugPreEnd();
	}
}

void	sqlrconnection::debugPreStart() {
	if (webdebug==-1) {
		char	*docroot=getenv("DOCUMENT_ROOT");
		if (docroot && docroot[0]) {
			webdebug=1;
		} else {
			webdebug=0;
		}
	}
	if (webdebug==1) {
		debugPrint("<pre>\n");
	}
}

void	sqlrconnection::debugPreEnd() {
	if (webdebug==1) {
		debugPrint("</pre>\n");
	}
}

void	sqlrconnection::debugPrintFunction(
				int (*printfunction)(const char *,...)) {
	this->printfunction=printfunction;
}

void	sqlrconnection::debugPrint(char *string) {
	if (printfunction) {
		(*printfunction)("%s",string);
	} else {
		cout << string;
	}
}

void	sqlrconnection::debugPrint(long number) {
	if (printfunction) {
		(*printfunction)("%d",number);
	} else {
		cout << number;
	}
}

void	sqlrconnection::debugPrint(double number) {
	if (printfunction) {
		(*printfunction)("%f",number);
	} else {
		cout << number;
	}
}

void	sqlrconnection::copyReferences() {

	// set the flag
	copyrefs=1;

	// make copies of some specific things
	if (server) {
		char *tempserver=new char[strlen(server)+1];
		strcpy(tempserver,server);
		server=tempserver;
	}
	if (listenerunixport) {
		char *templistenerunixport=new char[strlen(listenerunixport)+1];
		strcpy(templistenerunixport,listenerunixport);
		listenerunixport=templistenerunixport;
	}
	if (user) {
		char *tempuser=new char[strlen(user)+1];
		strcpy(tempuser,user);
		user=tempuser;
	}
	if (password) {
		char *temppassword=new char[strlen(password)+1];
		strcpy(temppassword,password);
		password=temppassword;
	}
}

void	sqlrconnection::clearSessionFlags() {

	// indicate that the session hasn't been suspended or ended
	endsessionsent=0;
	suspendsessionsent=0;
}

int	sqlrconnection::openSession() {

	if (debug) {
		debugPreStart();
		debugPrint("Connecting to listener...");
		debugPrint("\n");
		debugPreEnd();
	}

	// open a connection to the listener
	int	openresult=0;

	// first, try for a unix connection
	if (listenerunixport && listenerunixport[0]) {
		openresult=connectToServer(listenerunixport,
						retrytime,tries);
	}

	// then try for an inet connection
	if (!openresult && listenerinetport) {
		openresult=connectToServer(server,listenerinetport,
						retrytime,tries);
	}

	// handle failure to connect to listener
	if (!openresult) {
		setError("Couldn't connect to the listener.");
		return 0;
	}

	// authenticate with the listner
	if (authenticateWithListener()<1) {
		clientsocket::closeConnection();
		return 0;
	}

	// do we need to reconnect to the connection daemon
	getReconnect();


	if (reconnect==-1) {

		// if an error ocurred, set the error and close the connection
		setError("Failed to get whether we need to reconnect.\n A network error may have ocurred.");
		clientsocket::closeConnection();
		return 0;

	} else if (!reconnect) {

		// if we don't need to reconnect, just set the 
		// connected flag to true
		connected=1;

	} else if (reconnect==1) {

		// if we do need to reconnect, get which port(s) to reconnect
		// to and reconnect

		// try to get the connection daemon ports
		int	success=getNewPort();

		// close the connection to the listener
		clientsocket::closeConnection();

		// If getNewPort() returns -1 or 0 then an error ocurred.
		// If it returns 0, the error is already set.
		if (success<1) {
			if (success==-1) {
				setError("Failed to get connection ports.\n A network error may have ocurred.");
			}
			return 0;
		}

		if (debug) {
			debugPreStart();
			debugPrint("Reconnecting to ");
			debugPrint("\n");
			debugPrint("	unix port: ");
			debugPrint(connectionunixport);
			debugPrint("\n");
			debugPrint("	inet port: ");
			debugPrint((long)connectioninetport);
			debugPrint("\n");
			debugPreEnd();
		}

		// first, try for the unix port
		if (listenerunixport && listenerunixport[0] &&
					connectionunixport) {
			connected=connectToServer(connectionunixport,
						retrytime,tries);
		}

		// then try for the inet port
		if (!connected && connectioninetport) {
			connected=connectToServer(server,
						connectioninetport,
						retrytime,tries);
		}

		// did we successfully reconnect?
		if (connected) {

			if (debug) {
				debugPreStart();
				debugPrint("Connected.");
				debugPrint("\n");
				debugPreEnd();
			}

			clearSessionFlags();

			if (!authenticateWithConnection()) {
				return 0;
			}

		} else {

			// handle failure to connect to database 
			// connection daemon
			strstream	*errstr=new strstream();
			*errstr << "Couldn't connect to the database";
			*errstr << "connection daemon." << ENDL;
			if (connectionunixport) {
				*errstr << "Tried unix port ";
				*errstr << connectionunixport << ENDL;
			}
			if (connectioninetport) {
				*errstr << "Tried inet port ";
				*errstr << connectioninetport << ENDL;
			}
			*errstr << ENDS;
			setError(errstr->str());
			delete errstr;
			return 0;
		}
	}

	// if we made it here then everything went well;  listener 
	// authentication succeeded and we either didn't have to reconnect or
	// we sucessfully reconnected and sucessfully authenticated with the 
	// connection daemon
	return 1;
}

int	sqlrconnection::authenticateWithConnection() {

	writeData((unsigned short)AUTHENTICATE);
	if (debug) {
		debugPreStart();
		debugPrint("Authenticating with connection : ");
	}
	return genericAuthentication();
}

int	sqlrconnection::resumeSession(int port, char *socket) {

	// if already connected, end the session
	if (connected) {
		endSession();
	}

	// set the connectionunixport and connectioninetport values
	if (copyrefs) {
		if (strlen(socket)<=MAXPATHLEN) {
			strcpy(connectionunixportbuffer,socket);
			connectionunixport=connectionunixportbuffer;
		} else {
			connectionunixport="";
		}
	} else {
		connectionunixport=socket;
	}
	connectioninetport=(unsigned short)port;

	// first, try for the unix port
	if (socket && socket[0]) {
		connected=connectToServer(socket,retrytime,tries);
	}

	// then try for the inet port
	if (!connected) {
		connected=connectToServer(server,port,retrytime,tries);
	}

	if (debug) {
		debugPreStart();
		debugPrint("Resuming Session: ");
		debugPreEnd();
	}

	if (connected) {
		if (debug) {
			debugPreStart();
			debugPrint("success");
		       	debugPrint("\n");
			debugPreEnd();
		}
		clearSessionFlags();
	} else {
		if (debug) {
			debugPreStart();
			debugPrint("failure");
			debugPrint("\n");
			debugPreEnd();
		}
	}

	return connected;
}

void	sqlrconnection::closeConnection() {

	if (connected) {
		clientsocket::closeConnection();
		connected=0;
	}
}

sqlrcursor::sqlrcursor(sqlrconnection *sqlrc) {

	// copy references
	copyrefs=0;

	this->sqlrc=sqlrc;

	// put self in connection's cursor list
	if (sqlrc->lastcursor) {
		sqlrc->lastcursor->next=this;
		prev=sqlrc->lastcursor;
	} else {
		sqlrc->firstcursor=this;
		prev=NULL;
	}
	sqlrc->lastcursor=this;
	next=NULL;

	// session state
	cached=0;

	// query
	querybuffer=NULL;
	fullpath=NULL;

	// result set
	rsbuffersize=0;

	firstrowindex=0;
	rowcount=0;
	previousrowcount=0;
	actualrows=0;
	affectedrows=0;
	endofresultset=1;

	error=NULL;

	rows=NULL;
	extrarows=NULL;
	firstextrarow=NULL;
	rowstorage=NULL;
	fields=NULL;
	fieldlengths=NULL;
	getrowcount=0;
	getrowlengthcount=0;

	colcount=0;
	previouscolcount=0;
	columns=NULL;
	extracolumns=NULL;
	colstorage=NULL;
	columnnamearray=NULL;

	nonstandardtypes=NULL;

	returnnulls=0;

	// cache file
	cachesourcefd=-1;
	cachesourceindfd=-1;
	cachedestname=NULL;
	cachedestindname=NULL;
	cachedestfd=NULL;
	cachedestindfd=NULL;
	cacheon=0;

	// options...
	sendcolumninfo=SEND_COLUMN_INFO;

	// cursor id
	cursorid=0;

	initVariables();
}

sqlrcursor::~sqlrcursor() {

	// abort result set if necessary
	if (!endofresultset && sqlrc &&
			!sqlrc->endsessionsent && 
			!sqlrc->suspendsessionsent) {
		abortResultSet();
	}

	// deallocate copied references
	if (copyrefs) {
		deleteVariables();
	}

	// deallocate the query buffer
	if (querybuffer) {
		delete[] querybuffer;
	}

	// deallocate the fullpath (used for file queries)
	if (fullpath) {
		delete[] fullpath;
	}

	clearResultSet();
	if (columns) {
		delete[] columns;
	}
	if (extracolumns) {
		delete[] extracolumns;
	}
	if (colstorage) {
		delete[] colstorage;
	}
	if (rows) {
		for (int i=0; i<OPTIMISTIC_ROW_COUNT; i++) {
			delete rows[i];
		}
		delete[] rows;
	}
	if (rowstorage) {
		delete[] rowstorage;
	}

	// it's possible for the connection to be deleted before the 
	// cursor is, in that case, don't do any of this stuff
	if (sqlrc) {

		// remove self from connection's cursor list
		if (!next && !prev) {
			sqlrc->firstcursor=NULL;
			sqlrc->lastcursor=NULL;
		} else {
			sqlrcursor	*temp=next;
			if (next) {
				next->prev=prev;
			} else {
				sqlrc->lastcursor=prev;
			}
			if (prev) {
				prev->next=temp;
			} else {
				sqlrc->firstcursor=next;
			}
		}

		if (sqlrc->debug) {
			sqlrc->debugPreStart();
			sqlrc->debugPrint("Deallocated cursor\n");
			sqlrc->debugPreEnd();
		}
	}

	if (copyrefs && cachedestname) {
		delete[] cachedestname;
	}
	if (cachedestindname) {
		delete[] cachedestindname;
	}
}

void	sqlrcursor::copyReferences() {

	// set the flag
	copyrefs=1;
}

void	sqlrcursor::clearVariables() {

	// setting the bind/substitution variable 
	// counts to 0 effectively clears them
	subcount=0;
	inbindcount=0;
	outbindcount=0;
}

void	sqlrcursor::initVariables() {

	// initialize the bind and substitution variables
	for (int i=0; i<MAXVAR; i++) {
		inbindvars[i].variable=NULL;
		inbindvars[i].value.stringval=NULL;
		inbindvars[i].type=STRING_BIND;
		outbindvars[i].variable=NULL;
		outbindvars[i].value.stringval=NULL;
		outbindvars[i].type=STRING_BIND;
		subvars[i].variable=NULL;
		subvars[i].value.stringval=NULL;
		subvars[i].type=STRING_BIND;
	}
}

void	sqlrcursor::deleteVariables() {

	// if we were copying values, delete them
	for (int i=0; i<MAXVAR; i++) {
		if (inbindvars[i].variable) {
			delete[] inbindvars[i].variable;
		}
		if (inbindvars[i].type==STRING_BIND && 
			inbindvars[i].value.stringval) {
			delete[] inbindvars[i].value.stringval;
		}
		if (outbindvars[i].variable) {
			delete[] outbindvars[i].variable;
		}
		if (outbindvars[i].type==STRING_BIND && 
			outbindvars[i].value.stringval) {
			delete[] outbindvars[i].value.stringval;
		}
		if (subvars[i].variable) {
			delete[] subvars[i].variable;
		}
		if (subvars[i].type==STRING_BIND && 
			subvars[i].value.stringval) {
			delete[] subvars[i].value.stringval;
		}
	}
}

void	sqlrcursor::clearRows() {

	// delete linked list storing extra result set fields
	row	*currentrow;
	if (firstextrarow) {
		currentrow=firstextrarow;
		while (currentrow) {
			firstextrarow=currentrow->next;
			delete currentrow;
			currentrow=firstextrarow;
		}
		firstextrarow=NULL;
	}
	currentrow=NULL;

	// delete array pointing to linked list items
	if (extrarows) {
		delete[] extrarows;
		extrarows=NULL;
	}

	// delete arrays of fields and field lengths
	if (fields) {
		for (int i=0; i<rowcount; i++) {
			delete[] fields[i];
		}
		delete[] fields;
		fields=NULL;
	}
	if (fieldlengths) {
		for (int i=0; i<rowcount; i++) {
			delete[] fieldlengths[i];
		}
		delete[] fieldlengths;
		fieldlengths=NULL;
	}

	// reset the row storage pointer
	currentrowstorageptr=rowstorage;
}

void	sqlrcursor::clearCacheDest() {
	if (cachedestfd) {
		cachedestfd->close();
		cachedestindfd->close();
		delete cachedestfd;
		delete cachedestindfd;
		cachedestfd=NULL;
		cachedestindfd=NULL;
		cacheon=0;
	}
}

void	sqlrcursor::clearCacheSource() {
	if (cachesourcefd>-1) {
		close(cachesourcefd);
		cachesourcefd=-1;
	}
	if (cachesourceindfd>-1) {
		close(cachesourceindfd);
		cachesourceindfd=-1;
	}
}

void	sqlrcursor::clearError() {
	if (error) {
		delete[] error;
		error=NULL;
	}
	if (sqlrc) {
		sqlrc->clearError();
	}
}

int	sqlrcursor::fetchRowIntoBuffer(int row) {

	// if we getting the entire result set at once, then the result set 
	// buffer index is the requested row-firstrowindex
	if (!rsbuffersize) {
		if (row<rowcount && row>=firstrowindex) {
			return row-firstrowindex;
		}
		return -1;
	}

	// but, if we're not getting the entire result set at once
	// and if the requested row is not in the current range, 
	// fetch more data from the connection
	while (row>=(firstrowindex+rsbuffersize) && !endofresultset) {
		if (sqlrc->connected || 
				(cachesourcefd>-1 && cachesourceindfd>-1)) {
			clearRows();

			// if we're not fetching from a cached result set,
			// tell the server to send one 
			if (cachesourcefd==-1 && cachesourceindfd==-1) {
				sqlrc->writeData((unsigned short)
							FETCH_RESULT_SET);
				sqlrc->writeData(cursorid);
			}

			if (skipAndFetch(row)==-1 || parseData()==-1) {
				return -1;
			}
		} else {
			return -1;
		}
	}

	// return the buffer index corresponding to the requested row
	// or -1 if the requested row is past the end of the result set
	if (row<rowcount) {
		return row%rsbuffersize;
	}
	return -1;
}

char	**sqlrcursor::getColumnNames() {

	if (sendcolumninfo==DONT_SEND_COLUMN_INFO ||
			sentcolumninfo==DONT_SEND_COLUMN_INFO) {
		return NULL;
	}

	if (!columnnamearray) {
		if (sqlrc->debug) {
			sqlrc->debugPreStart();
			sqlrc->debugPrint("Creating Column Arrays...\n");
			sqlrc->debugPreEnd();
		}
	
		// build a 2d array of pointers to the column names
		columnnamearray=new char *[colcount+1];
		columnnamearray[colcount]=NULL;
		for (int i=0; i<colcount; i++) {
			columnnamearray[i]=getColumn(i)->name;
		}
	}
	return columnnamearray;
}

char	*sqlrcursor::getColumnName(int col) {

	if (sendcolumninfo==SEND_COLUMN_INFO && 
			sentcolumninfo==SEND_COLUMN_INFO &&
			colcount && col>=0 && col<colcount) {
		return getColumn(col)->name;
	}
	return NULL;
}

char	*sqlrcursor::getColumnType(int col) {

	if (sendcolumninfo==SEND_COLUMN_INFO && 
			sentcolumninfo==SEND_COLUMN_INFO &&
			colcount && col>=0 && col<colcount) {
		column	*whichcolumn=getColumn(col);
		if (whichcolumn->type>=NUMBER_OF_DATATYPES) {
			return nonstandardtypes[col];
		}
		return datatypestring[whichcolumn->type];
	}
	return NULL;
}

int	sqlrcursor::getColumnLength(int col) {

	if (sendcolumninfo==SEND_COLUMN_INFO && 
			sentcolumninfo==SEND_COLUMN_INFO &&
			colcount && col>=0 && col<colcount) {
		return getColumn(col)->length;
	}
	return 0;
}

int	sqlrcursor::getLongest(int col) {

	if (sendcolumninfo==SEND_COLUMN_INFO && 
			sentcolumninfo==SEND_COLUMN_INFO &&
			colcount && col>=0 && col<colcount) {
		return getColumn(col)->longest;
	} 
	return 0;
}

char	*sqlrcursor::getColumnType(char *col) {

	if (sendcolumninfo==SEND_COLUMN_INFO && 
			sentcolumninfo==SEND_COLUMN_INFO) {
		column	*whichcolumn;
		for (int i=0; i<colcount; i++) {
			whichcolumn=getColumn(i);
			if (!strcasecmp(whichcolumn->name,col)) {
				if (whichcolumn->type>=NUMBER_OF_DATATYPES) {
					return nonstandardtypes[i];
				}
				return datatypestring[whichcolumn->type];
			}
		}
	}
	return NULL;
}

int	sqlrcursor::getColumnLength(char *col) {

	if (sendcolumninfo==SEND_COLUMN_INFO && 
			sentcolumninfo==SEND_COLUMN_INFO) {
		column	*whichcolumn;
		for (int i=0; i<colcount; i++) {
			whichcolumn=getColumn(i);
			if (!strcasecmp(whichcolumn->name,col)) {
				return whichcolumn->length;
			}
		}
	}
	return 0;
}

int	sqlrcursor::getLongest(char *col) {

	if (sendcolumninfo==SEND_COLUMN_INFO && 
			sentcolumninfo==SEND_COLUMN_INFO) {
		column	*whichcolumn;
		for (int i=0; i<colcount; i++) {
			whichcolumn=getColumn(i);
			if (!strcasecmp(whichcolumn->name,col)) {
				return whichcolumn->longest;
			}
		}
	}
	return 0;
}

void	sqlrcursor::cacheData() {

	if (!cachedestfd) {
		return;
	}

	// write the data to the cache file
	for (int i=0; i<rowcount-firstrowindex; i++) {

		// write the row offset to the index
		cachedestindfd->seekp(13+sizeof(long)+
					((firstrowindex+i)*sizeof(long)));
		long	position=(long)cachedestfd->tellp();
		cachedestindfd->write((char *)&position,sizeof(long));

		// write the row to the cache file
		for (int j=0; j<colcount; j++) {
			unsigned short	type;
			long	len;
			char	*field=getFieldInternal(i,j);
			if (field) {
				type=NORMAL_DATA;
				len=strlen(field);
				cachedestfd->write((char *)&type,
							sizeof(unsigned short));
				cachedestfd->write((char *)&len,
							sizeof(long));
				if (len>0) {
					cachedestfd->write(
						(char *)field,strlen(field));
				}
			} else {
				type=NULL_DATA;
				cachedestfd->write((char *)&type,
							sizeof(unsigned short));
			}
		}
	}

	if (endofresultset) {
		finishCaching();
	}
}

int	sqlrcursor::firstRowIndex() {
	return firstrowindex;
}

int	sqlrcursor::endOfResultSet() {
	return endofresultset;
}

int	sqlrcursor::rowCount() {
	return rowcount;
}

int	sqlrcursor::affectedRows() {
	if (knowsaffectedrows==AFFECTED_ROWS) {
		return affectedrows;
	}
	return -1;
}

int	sqlrcursor::totalRows() {
	if (knowsactualrows==ACTUAL_ROWS) {
		return actualrows;
	}
	return -1;
}

int	sqlrcursor::colCount() {
	return colcount;
}

char	*sqlrcursor::errorMessage() {

	if (error) {
		return error;
	} else if (sqlrc->error) {
		return sqlrc->error;
	}
	return NULL;
}

void	sqlrcursor::getNullsAsEmptyStrings() {
	returnnulls=0;
}

void	sqlrcursor::getNullsAsNulls() {
	returnnulls=1;
}

char	*sqlrcursor::getField(int row, int col) {

	if (rowcount && row>=0 && row>=firstrowindex && 
					col>=0 && col<colcount) {

		// in the event that we're stepping through the result set 
		// instead of buffering the entire thing, the requested row
		// may have to be fetched into the buffer...
		int	whichrow=fetchRowIntoBuffer(row);

		if (whichrow>-1) {
			char	*retval=getFieldInternal(whichrow,col);
			return retval;
		}
	}
	return NULL;
}

char	*sqlrcursor::getField(int row, char *col) {

	if (sendcolumninfo==SEND_COLUMN_INFO && 
			sentcolumninfo==SEND_COLUMN_INFO &&
			rowcount && row>=0 && row>=firstrowindex) {
		for (int i=0; i<colcount; i++) {
			if (!strcasecmp(getColumn(i)->name,col)) {

				// in the event that we're stepping through the
				// result set instead of buffering the entire 
				// thing, the requested row may have to be 
				// fetched into the buffer...
				int	whichrow=fetchRowIntoBuffer(row);

				if (whichrow>-1) {
					return getFieldInternal(whichrow,i);
				}
				return NULL;
			}
		}
	}
	return NULL;
}

long	sqlrcursor::getFieldLength(int row, int col) {

	if (rowcount && row>=0 && row>=firstrowindex && 
					col>=0 && col<colcount) {

		// in the event that we're stepping through the result set 
		// instead of buffering the entire thing, the requested row
		// may have to be fetched into the buffer...
		int	whichrow=fetchRowIntoBuffer(row);

		if (whichrow>-1) {
			return getFieldLengthInternal(whichrow,col);
		}
	}
	return -1;
}

long	sqlrcursor::getFieldLength(int row, char *col) {

	if (sendcolumninfo==SEND_COLUMN_INFO && 
			sentcolumninfo==SEND_COLUMN_INFO &&
			rowcount && row>=0 && row>=firstrowindex) {

		for (int i=0; i<colcount; i++) {
			if (!strcasecmp(getColumn(i)->name,col)) {

				// in the event that we're stepping through the
				// result set instead of buffering the entire 
				// thing, the requested row may have to be 
				// fetched into the buffer...
				int	whichrow=fetchRowIntoBuffer(row);

				if (whichrow>-1) {
					return getFieldLengthInternal(whichrow,
									i);
				}
				return -1;
			}
		}
	}
	return -1;
}

char	**sqlrcursor::getRow(int row) {

	if (rowcount && row>=0 && row>=firstrowindex) {

		// in the event that we're stepping through the result set 
		// instead of buffering the entire thing, the requested row
		// may have to be fetched into the buffer...
		int	whichrow=fetchRowIntoBuffer(row);

		if (whichrow>-1) {
			if (!fields) {
				createFields();
			}
			return fields[row];
		}
	}
	return NULL;
}

long	*sqlrcursor::getRowLengths(int row) {

	if (rowcount && row>=0 && row>=firstrowindex) {

		// in the event that we're stepping through the result set 
		// instead of buffering the entire thing, the requested row
		// may have to be fetched into the buffer...
		int	whichrow=fetchRowIntoBuffer(row);

		if (whichrow>-1) {
			if (!fieldlengths) {
				createFieldLengths();
			}
			return (long *)fieldlengths[whichrow];
		}
	}
	return NULL;
}

void	sqlrcursor::createExtraRowArray() {

	// create the arrays
	int	howmany=rowcount-firstrowindex-OPTIMISTIC_ROW_COUNT;
	extrarows=new row *[howmany];
	
	// populate the arrays
	row	*currentrow=firstextrarow;
	for (int i=0; i<howmany; i++) {
		extrarows[i]=currentrow;
		currentrow=currentrow->next;
	}
}

void	sqlrcursor::createFields() {
	fields=new char **[rowcount+1];
	fields[rowcount]=(char **)NULL;
	for (int i=0; i<rowcount; i++) {
		fields[i]=new char *[colcount+1];
		fields[i][colcount]=(char *)NULL;
		for (int j=0; j<colcount; j++) {
			fields[i][j]=getFieldInternal(i,j);
		}
	}
}

void	sqlrcursor::createFieldLengths() {
	fieldlengths=new unsigned long *[rowcount+1];
	fieldlengths[rowcount]=(unsigned long)NULL;
	for (int i=0; i<rowcount; i++) {
		fieldlengths[i]=new unsigned long[colcount+1];
		fieldlengths[i][colcount]=(unsigned long)NULL;
		for (int j=0; j<colcount; j++) {
			fieldlengths[i][j]=getFieldLengthInternal(i,j);
		}
	}
}

void	sqlrcursor::setError(char *err) {

	if (sqlrc->debug) {
		sqlrc->debugPreStart();
		sqlrc->debugPrint("Setting Error\n");
		sqlrc->debugPreEnd();
	}

	error=new char[strlen(err)+1];
	strcpy(error,err);

	handleError();
}

void	sqlrcursor::getErrorFromServer() {

	if (sqlrc->debug) {
		sqlrc->debugPreStart();
		sqlrc->debugPrint("Getting Error From Server\n");
		sqlrc->debugPreEnd();
	}

	// get the length of the error string
	unsigned short	length;
	if (getShort(&length)!=sizeof(unsigned short)) {
		error=new char[55];
		strcpy(error,"There was an error, but the connection died trying to retrieve it.  Sorry.");
	} else {
		// get the error string
		error=new char[length+1];
		sqlrc->readData(error,length);
		error[length]=(char)NULL;
	}
	
	handleError();
}

void	sqlrcursor::handleError() {

	if (sqlrc->debug) {
		sqlrc->debugPreStart();
		sqlrc->debugPrint(error);
		sqlrc->debugPrint("\n");
		sqlrc->debugPreEnd();
	}

	endofresultset=1;

	if (resumed || !cachedestfd) {
		return;
	}

	// write the number of returned rows, affected rows 
	// and a zero to terminate the column descriptions
	unsigned short	val=NO_ACTUAL_ROWS;
	cachedestfd->write((char *)&val,sizeof(unsigned short));
	val=NO_AFFECTED_ROWS;
	cachedestfd->write((char *)&val,sizeof(unsigned short));
	val=END_COLUMN_INFO;
	cachedestfd->write((char *)&val,sizeof(unsigned short));
	finishCaching();
}

int	sqlrcursor::skipAndFetch(int rowtoget) {

	if (sqlrc->debug) {
		sqlrc->debugPreStart();
		sqlrc->debugPrint("Skipping and Fetching\n");
		if (rowtoget>-1) {
			sqlrc->debugPrint("	row to get: ");
			sqlrc->debugPrint((long)rowtoget);
			sqlrc->debugPrint("\n");
		}
		sqlrc->debugPreEnd();
	}

	// if we're stepping through the result set, we can possibly 
	// skip a big chunk of it...
	if (!skipRows(rowtoget)) {
		return -1;
	}

	// tell the connection how many rows to send
	fetchRows();
	return 1;
}

void	sqlrcursor::fetchRows() {

	if (sqlrc->debug) {
		sqlrc->debugPreStart();
		sqlrc->debugPrint("Fetching ");
		sqlrc->debugPrint(rsbuffersize);
		sqlrc->debugPrint(" rows\n");
		sqlrc->debugPreEnd();
	}

	// if we're reading from a cached result set, do nothing
	if (cachesourcefd>-1 && cachesourceindfd>-1) {
		return;
	}

	// otherwise, send to the connection the number of rows to send back
	sqlrc->writeData((unsigned long)rsbuffersize);
}

int	sqlrcursor::skipRows(int rowtoget) {

	// if we're reading from a cached result set we have to manually skip
	if (cachesourcefd>-1 && cachesourceindfd>-1) {

		// if rowtoget is -1 then don't skip,
		// otherwise skip to the next block of rows
		if (rowtoget==-1) {
			return 1;
		} else {
			rowcount=rowtoget-(rowtoget%rsbuffersize);
		}

		// get the row offset from the index
		lseek(cachesourceindfd,
			13+sizeof(long)+(rowcount*sizeof(long)),
			SEEK_SET);
		long	rowoffset;
		if (read(cachesourceindfd,&rowoffset,sizeof(long))!=
								sizeof(long)) {

			setError("The cache file index appears to be corrupt.");
			return 0;
		}

		// skip to that offset in the cache file
		lseek(cachesourcefd,rowoffset,SEEK_SET);
		return 1;
	}

	// calculate how many rows to skip unless we're buffering the entire
	// result set or caching the result set
	unsigned long	skip=0;
	if (rsbuffersize && !cachedestfd) {
		skip=(long)((rowtoget-(rowtoget%rsbuffersize))-rowcount); 
		rowcount=rowcount+skip;
	}
	if (sqlrc->debug) {
		sqlrc->debugPreStart();
		sqlrc->debugPrint("Skipping ");
		sqlrc->debugPrint((long)skip);
		sqlrc->debugPrint(" rows\n");
		sqlrc->debugPreEnd();
	}

	// if we're reading from a connection, send the connection the 
	// number of rows to skip
	sqlrc->writeData(skip);
	return 1;
}

int	sqlrcursor::parseData() {

	if (sqlrc->debug) {
		sqlrc->debugPreStart();
		sqlrc->debugPrint("Parsing Data\n");
		sqlrc->debugPreEnd();
	}

	// if we're already at the end of the result set, then just return
	if (endofresultset) {
		if (sqlrc->debug) {
			sqlrc->debugPreStart();
			sqlrc->debugPrint("Already at the end of the result set\n");
			sqlrc->debugPreEnd();
		}
		return 1;
	}

	// useful variables
	unsigned short	type;
	unsigned long	length;
	char	*buffer=NULL;
	long	colindex=0;
	column	*currentcol;
	row	*currentrow=NULL;

	// set firstrowindex to the index of the first row in the buffer
	firstrowindex=rowcount;

	// keep track of how large the buffer is
	int	rsbuffercount=0;

	// get rows
	for (;;) {

		// get the type of the field
		if (getShort(&type)!=sizeof(unsigned short)) {
			return -1;
		}

		// check for the end of the result set
		if (type==END_RESULT_SET) {

			if (sqlrc->debug) {
				sqlrc->debugPreStart();
				sqlrc->debugPrint("Got end of result set.\n");
				sqlrc->debugPreEnd();
			}
			endofresultset=1;

			// if we were stepping through a cached result set
			// then we need to close the file
			clearCacheSource();
			break;
		} 

		// if we're on the first column, start a new row,
		// reset the column pointer, and increment the
		// buffer counter and total row counter
		if (colindex==0) {

			if (rowcount<OPTIMISTIC_ROW_COUNT) {
				if (!rows) {
					createRowBuffers();
				}
				currentrow=rows[rsbuffercount];
			} else {
				if (!firstextrarow) {
					currentrow=new row(colcount);
					firstextrarow=currentrow;
				} else {
					currentrow->next=new row(colcount);
					currentrow=currentrow->next;
				}
			}
			if (colcount>previouscolcount) {
				currentrow->resize(colcount);
			}

			rsbuffercount++;
			rowcount++;
		}

		if (type==NULL_DATA) {

			// handle null data
			if (returnnulls) {
				buffer=NULL;
			} else {
				buffer=getRowStorage(1);
				buffer[0]=(char)NULL;
			}
			length=0;

		} else if (type==NORMAL_DATA) {
		
			// handle non-null data
			if (getLong(&length)!=sizeof(unsigned long)) {
				return -1;
			}

			// for non-long, non-NULL datatypes...
			// get the field into a buffer
			buffer=getRowStorage(length+1);
			if (getString(buffer,length)!=length) {
				return -1;
			}
			buffer[length]=(char)NULL;

		} else if (type==START_LONG_DATA) {

			// handle a long datatype
			char	*oldbuffer;
			int	totallength=0;
			while (1) {

				// get the type of the chunk
				if (getShort(&type)!=sizeof(unsigned short)) {
					return -1;
				}

				// check to see if we're done
				if (type==END_LONG_DATA) {
					break;
				}

				// get the length of the chunk
				if (getLong(&length)!=sizeof(unsigned long)) {
					if (buffer) {
						delete[] buffer;
					}
					return -1;
				}

				buffer=new char[totallength+length+1];
				if (totallength) {
					strncpy(buffer,oldbuffer,totallength);
					delete[] oldbuffer;
					oldbuffer=buffer;
					buffer=buffer+totallength;
				} else {
					oldbuffer=buffer;
				}
				totallength=totallength+length;

				if (getString(buffer,length)!=length) {
					delete[] buffer;
					return -1;
				}
				buffer[length]=(char)NULL;
			}
			buffer=oldbuffer;
			length=totallength;

		}

		// add the buffer to the current row
		currentrow->addField(colindex,buffer,length);
	
		if (sqlrc->debug) {
			sqlrc->debugPreStart();
			if (buffer) {
				sqlrc->debugPrint("\"");
				sqlrc->debugPrint(buffer);
				sqlrc->debugPrint("\",");
			} else {
				sqlrc->debugPrint(buffer);
				sqlrc->debugPrint(",");
			}
			sqlrc->debugPreEnd();
		}

		// keep track of the longest field
		if (sendcolumninfo==SEND_COLUMN_INFO && 
				sentcolumninfo==SEND_COLUMN_INFO) {
			currentcol=getColumn(colindex);
			if (length>currentcol->longest) {
				currentcol->longest=length;
			}
		}

		// move to the next column, handle end of row 
		colindex++;
		if (colindex==colcount) {

			colindex=0;

			if (sqlrc->debug) {
				sqlrc->debugPreStart();
				sqlrc->debugPrint("\n");
				sqlrc->debugPreEnd();
			}

			// check to see if we've gotten enough rows
			if (rsbuffersize && 
				rsbuffercount==rsbuffersize) {
				break;
			}
		}
	}

	// terminate the row list
	if (rowcount>=OPTIMISTIC_ROW_COUNT && currentrow) {
		currentrow->next=NULL;
		createExtraRowArray();
	}

	// cache the rows
	cacheData();

	return 1;
}

int	sqlrcursor::parseOutputBinds() {

	if (sqlrc->debug) {
		sqlrc->debugPreStart();
		sqlrc->debugPrint("Receiving Output Bind Values: \n");
		sqlrc->debugPreEnd();
	}

	// useful variables
	unsigned short	type;
	unsigned short	length;
	int	count=0;

	// get the bind values
	for (;;) {

		// get the data type
		if (getShort(&type)!=sizeof(unsigned short)) {
			return -1;
		}

		// check for end of bind values
		if (type==END_BIND_VARS) {
			break;
		} else if (type==NULL_DATA) {
			// handle a null value
			if (returnnulls) {
				outbindvars[count].value.stringval=NULL;
			} else {
				outbindvars[count].value.stringval=new char[1];
				outbindvars[count].value.stringval[0]=
								(char)NULL;
			}
		} else {
			// get the value length
			if (getShort(&length)!=sizeof(unsigned short)) {
				return -1;
			}
			outbindvars[count].value.stringval=new char[length+1];
			if (getString(outbindvars[count].value.
						stringval,length)!=length) {
				return -1;
			}
			outbindvars[count].value.stringval[length]=(char)NULL;
		}

		if (sqlrc->debug) {
			sqlrc->debugPreStart();
			sqlrc->debugPrint(outbindvars[count].variable);
			sqlrc->debugPrint("=");
			sqlrc->debugPrint(outbindvars[count].value.stringval);
			sqlrc->debugPrint("\n");
			sqlrc->debugPreEnd();
		}

		count++;
	}

	// cache the output binds
	cacheOutputBinds(count);

	return 1;
}

void	sqlrcursor::cacheNoError() {

	if (resumed || !cachedestfd) {
		return;
	}

	// write the number of returned rows
	unsigned short	noerror=NO_ERROR;
	cachedestfd->write((char *)&noerror,sizeof(unsigned short));
}

void	sqlrcursor::cacheColumnInfo() {

	if (resumed || !cachedestfd) {
		return;
	}

	// write the number of returned rows
	cachedestfd->write((char *)&knowsactualrows,sizeof(unsigned short));
	if (knowsactualrows==ACTUAL_ROWS) {
		cachedestfd->write((char *)&actualrows,sizeof(unsigned long));
	}

	// write the number of affected rows
	cachedestfd->write((char *)&knowsaffectedrows,sizeof(unsigned short));
	if (knowsaffectedrows==AFFECTED_ROWS) {
		cachedestfd->write((char *)&affectedrows,sizeof(unsigned long));
	}

	// write whether or not the column info is is cached
	cachedestfd->write((char *)&sentcolumninfo,sizeof(unsigned short));

	// write the column count
	cachedestfd->write((char *)&colcount,sizeof(unsigned long));

	// write column descriptions to the cache file
	if (sendcolumninfo==SEND_COLUMN_INFO && 
			sentcolumninfo==SEND_COLUMN_INFO) {

		unsigned short	namelen;
		column	*whichcolumn;
		for (int i=0; i<colcount; i++) {
			whichcolumn=getColumn(i);
			namelen=strlen(whichcolumn->name);
			cachedestfd->write((char *)&namelen,
						sizeof(unsigned short));
			cachedestfd->write(whichcolumn->name,namelen);
			cachedestfd->write((char *)&whichcolumn->type,
						sizeof(unsigned short));
			cachedestfd->write((char *)&whichcolumn->length,
						sizeof(unsigned long));
		}
	}
}

void	sqlrcursor::cacheOutputBinds(int count) {

	if (resumed || !cachedestfd) {
		return;
	}

	// write the variable/value pairs to the cache file
	unsigned short	len;
	unsigned short	type;
	for (int i=0; i<count; i++) {

		if (outbindvars[i].value.stringval) {
			type=STRING_BIND;
		} else {
			type=NULL_BIND;
		}
		cachedestfd->write((char *)&type,sizeof(unsigned short));

		len=strlen(outbindvars[i].variable);
		cachedestfd->write((char *)&len,sizeof(unsigned short));
		cachedestfd->write(outbindvars[i].variable,len);

		len=strlen(outbindvars[i].value.stringval);
		cachedestfd->write((char *)&len,sizeof(unsigned short));
		cachedestfd->write((char *)&outbindvars[i].value.stringval,
									len);
	}

	// terminate the list of output binds
	unsigned short	endbinds=END_BIND_VARS;
	cachedestfd->write((char *)&endbinds,sizeof(unsigned short));
}

void	sqlrcursor::suspendCaching() {

	if (!cachedestfd) {
		return ;
	}

	// close the cache file and clean up
	clearCacheDest();
}

int	sqlrcursor::noError() {

	if (sqlrc->debug) {
		sqlrc->debugPreStart();
		sqlrc->debugPrint("Checking For An Error... ");
		sqlrc->debugPreEnd();
	}

	// get a flag indicating whether there's been an error or not
	unsigned short	success;
	if (getShort(&success)!=sizeof(unsigned short)) {
		return -1;
	}
	if (success==NO_ERROR) {
		if (sqlrc->debug) {
			sqlrc->debugPreStart();
			sqlrc->debugPrint("none.\n");
			sqlrc->debugPreEnd();
		}
		cacheNoError();
		return 1;
	} else if (success==ERROR) {
		if (sqlrc->debug) {
			sqlrc->debugPreStart();
			sqlrc->debugPrint("error!!!\n");
			sqlrc->debugPreEnd();
		}
		return 0;
	}
	return -1;
}

int	sqlrcursor::getCursorId() {

	if (sqlrc->debug) {
		sqlrc->debugPreStart();
		sqlrc->debugPrint("Getting Cursor ID...\n");
		sqlrc->debugPreEnd();
	}
	if (sqlrc->readData(&cursorid)!=sizeof(unsigned short)) {
		return 0;
	}
	if (sqlrc->debug) {
		sqlrc->debugPreStart();
		sqlrc->debugPrint("Cursor ID: ");
		sqlrc->debugPrint((long)cursorid);
		sqlrc->debugPrint("\n");
		sqlrc->debugPreEnd();
	}
	return 1;
}

int	sqlrcursor::getSuspended() {

	// see if the result set of that cursor is actually suspended
	unsigned short	suspendedresultset;
	if (sqlrc->readData(&suspendedresultset)!=
					sizeof(unsigned short)) {
		return -1;
	}

	if (suspendedresultset==SUSPENDED_RESULT_SET) {

		// If it was suspended the server will send the index of the 
		// last row from the previous result set.
		// Initialize firstrowindex and rowcount from this index.
		sqlrc->readData(&firstrowindex);
		rowcount=firstrowindex+1;
	
		if (sqlrc->debug) {
			sqlrc->debugPreStart();
			sqlrc->debugPrint("Previous result set was ");
	       		sqlrc->debugPrint("suspended at row index: ");
			sqlrc->debugPrint((long)firstrowindex);
			sqlrc->debugPrint("\n");
			sqlrc->debugPreEnd();
		}
		return 1;
	} else if (suspendedresultset==NO_SUSPENDED_RESULT_SET) {
		if (sqlrc->debug) {
			sqlrc->debugPreStart();
			sqlrc->debugPrint("Previous result set was ");
	       		sqlrc->debugPrint("not suspended.\n");
			sqlrc->debugPreEnd();
		}
		return 1;
	}
	return 0;
}

int	sqlrcursor::parseColumnInfo() {

	if (sqlrc->debug) {
		sqlrc->debugPreStart();
		sqlrc->debugPrint("Parsing Column Info\n");
		sqlrc->debugPreEnd();
	}

	if (sqlrc->debug) {
		sqlrc->debugPreStart();
		sqlrc->debugPrint("Actual row count: ");
		sqlrc->debugPreEnd();
	}

	// first get whether the server knows the total number of rows or not
	if (getShort(&knowsactualrows)!=sizeof(unsigned short)) {
		return -1;
	}

	// get the number of rows returned by the query
	if (knowsactualrows==ACTUAL_ROWS) {
		if (getLong(&actualrows)!=sizeof(unsigned long)) {
			return -1;
		}
		if (sqlrc->debug) {
			sqlrc->debugPreStart();
			sqlrc->debugPrint((long)actualrows);
			sqlrc->debugPreEnd();
		}
	} else {
		if (sqlrc->debug) {
			sqlrc->debugPreStart();
			sqlrc->debugPrint("unknown");
			sqlrc->debugPreEnd();
		}
	}

	if (sqlrc->debug) {
		sqlrc->debugPreStart();
		sqlrc->debugPrint("\n");
		sqlrc->debugPrint("Affected row count: ");
		sqlrc->debugPreEnd();
	}

	// get whether the server knows the number of affected rows or not
	if (getShort(&knowsaffectedrows)!=sizeof(unsigned short)) {
		return -1;
	}

	// get the number of rows affected by the query
	if (knowsaffectedrows==AFFECTED_ROWS) {
		if (getLong(&affectedrows)!=sizeof(unsigned long)) {
			return -1;
		}
		if (sqlrc->debug) {
			sqlrc->debugPreStart();
			sqlrc->debugPrint((long)affectedrows);
			sqlrc->debugPreEnd();
		}
	} else {
		if (sqlrc->debug) {
			sqlrc->debugPreStart();
			sqlrc->debugPrint("unknown");
			sqlrc->debugPreEnd();
		}
	}

	if (sqlrc->debug) {
		sqlrc->debugPreStart();
		sqlrc->debugPrint("\n");
		sqlrc->debugPreEnd();
	}

	// get whether the server is sending column info or not
	if (getShort(&sentcolumninfo)!=sizeof(unsigned short)) {
		return -1;
	}

	// get column count
	if (getLong(&colcount)!=sizeof(unsigned long)) {
		return -1;
	}
	if (sqlrc->debug) {
		sqlrc->debugPreStart();
		sqlrc->debugPrint("Column count: ");
		sqlrc->debugPrint((long)colcount);
		sqlrc->debugPrint("\n");
		sqlrc->debugPreEnd();
	}

	if (sendcolumninfo==SEND_COLUMN_INFO && 
			sentcolumninfo==SEND_COLUMN_INFO) {

		createColumnBuffers();

		// some useful variables
		unsigned short	length;
		column		*currentcol;

		// get the columninfo segment
		for (long i=0; i<colcount; i++) {
	
			// get the column name length
			if (getShort(&length)!=sizeof(unsigned short)) {
				return -1;
			}
	
			// which column to use
			currentcol=getColumn(i);
	
			// get the column name
			currentcol->name=getColStorage(length+1);
			if (getString(currentcol->name,length)!=length) {
				return -1;
			}
			currentcol->name[length]=(char)NULL;

			// get the column type
			if (getShort(&currentcol->type)!=
						sizeof(unsigned short)) {
				return -1;
			}

			// handle non-standard column numbers (for example,
			// from postgresql)
			if (currentcol->type>=NUMBER_OF_DATATYPES) {
				if (!nonstandardtypes) {
					nonstandardtypescount=colcount;
					nonstandardtypes=new char *[colcount];
					for (int j=0; j<colcount; j++) {
						nonstandardtypes[j]=new char[6];
					}
				}
				sprintf(nonstandardtypes[i],"%d",
					currentcol->type-NUMBER_OF_DATATYPES);
			}

			// get the column length
			if (getLong(&currentcol->length)!=
						sizeof(unsigned long)) {
				return -1;
			}

			// initialize the longest value
			currentcol->longest=0;
	
			if (sqlrc->debug) {
				sqlrc->debugPreStart();
				sqlrc->debugPrint("\"");
				sqlrc->debugPrint(currentcol->name);
				sqlrc->debugPrint("\",");
				sqlrc->debugPrint("\"");
				if (currentcol->type>NUMBER_OF_DATATYPES) {
					sqlrc->debugPrint(nonstandardtypes[i]);
				} else {
					sqlrc->debugPrint(datatypestring[
							currentcol->type]);
				}
				sqlrc->debugPrint("\",");
				sqlrc->debugPrint("\"");
				sqlrc->debugPrint((long)currentcol->length);
				sqlrc->debugPrint("\"");
				sqlrc->debugPrint("\n");
				sqlrc->debugPreEnd();
			}

		}
	}

	// cache the column definitions
	cacheColumnInfo();

	return 1;
}

void	sqlrcursor::createColumnBuffers() {

	// we could get really sophisticated here and keep stats on the number
	// of columns that previous queries returned and adjust the size of
	// "columns" periodically, but for now, we'll just use a static size

	// create the standard set of columns, this will hang around until
	// the cursor is deleted
	if (!columns) {
		columns=new column[OPTIMISTIC_COLUMN_COUNT];
		colstorage=new char[OPTIMISTIC_COLUMN_DATA_SIZE];
		currentcolstorageptr=colstorage;
		endcolstorageptr=colstorage+OPTIMISTIC_COLUMN_DATA_SIZE;
	}

	// if there are more columns than our static column buffer
	// can handle, create extra columns, these will be deleted after each
	// query
	if (colcount>OPTIMISTIC_COLUMN_COUNT && colcount>previouscolcount) {
		if (extracolumns) {
			delete[] extracolumns;
		}
		extracolumns=new column[colcount-OPTIMISTIC_COLUMN_COUNT];
	}
}

column	*sqlrcursor::getColumn(int index) {
	if (index<OPTIMISTIC_COLUMN_COUNT) {
		return &columns[index];
	}
	return &extracolumns[index-OPTIMISTIC_COLUMN_COUNT];
}

void	sqlrcursor::abortResultSet() {

	// If the end of the previous result set was never reached, abort it.
	// If we're caching data to a local file, get the rest of the data; we
	// won't have to abort the result set in that case, the server will
	// do it.
	if (sqlrc->debug) {
		sqlrc->debugPreStart();
		sqlrc->debugPrint("Aborting Result Set For Cursor: ");
		sqlrc->debugPrint((long)cursorid);
		sqlrc->debugPrint("\n");
		sqlrc->debugPreEnd();
	}

	if (sqlrc->connected || cached) {
		if (cachedestfd && cachedestindfd) {
			if (sqlrc->debug) {
				sqlrc->debugPreStart();
				sqlrc->debugPrint("Getting the rest of the result set, since this is a cached result set.\n");
				sqlrc->debugPreEnd();
			}
			while (!endofresultset) {
				clearRows();

				// if we're not fetching from a cached result 
				// set tell the server to send one 
				if (cachesourcefd==-1 && cachesourceindfd==-1) {
					sqlrc->writeData((unsigned short)
							FETCH_RESULT_SET);
					sqlrc->writeData(cursorid);
				}

				// parseData should call finishCaching when
				// it hits the end of the result set, but
				// if it or skipAndFetch return a -1 (network
				// error) we'll have to call it ourselves.
				if (skipAndFetch(-1)==-1 || parseData()==-1) {
					finishCaching();
					return;
				}
			}
		} else {
			sqlrc->writeData((unsigned short)ABORT_RESULT_SET);
			sqlrc->writeData(cursorid);
		}
	}
}

void	sqlrcursor::suspendResultSet() {

	if (sqlrc->debug) {
		sqlrc->debugPreStart();
		sqlrc->debugPrint("Suspending Result Set\n");
		sqlrc->debugPreEnd();
	}
	if (sqlrc->connected && !cached) {
		sqlrc->writeData((unsigned short)SUSPEND_RESULT_SET);
		sqlrc->writeData(cursorid);
	}
	suspendCaching();
	suspendresultsetsent=1;
}

int	sqlrcursor::processResultSet(int rowtoget) {

	// start caching the result set
	if (cacheon) {
		startCaching();
	}

	// parse the columninfo and data
	int	success=1;

	// skip and fetch here if we're not reading from a cached result set
	// this way, everything gets done in 1 round trip
	if (cachesourcefd==-1) {
		success=skipAndFetch(firstrowindex+rowtoget);
	}

	// get data back from the server
	if (success>0 && (success=noError())>0 &&
			((cachesourcefd>-1 && cachesourceindfd>-1) ||
			((cachesourcefd==-1&& cachesourceindfd==-1)  && 
				(success=getCursorId()) && 
				(success=getSuspended())>0)) &&
			(success=parseColumnInfo())>0 && 
			(success=parseOutputBinds())>0) {

		// skip and fetch here if we're reading from a cached result set
		if (cachesourcefd>-1) {
			success=skipAndFetch(firstrowindex+rowtoget);
		}

		// parse the data
		if (success>-1) {
			success=parseData();
		}
	}

	// handle error responses
	if (success==0) {
		getErrorFromServer();
		return 0;
	} else if (success==-1) {
		clearResultSet();
		setError("Failed to execute the query and/or process the result set.\n A query, bind variable or bind value could be too large, there could be too \n many bind variables, or a network error may have ocurred.");
		sqlrc->endSession();
		return 0;
	}
	return 1;
}

void	sqlrcursor::startCaching() {

	if (!resumed) {
		if (sqlrc->debug) {
			sqlrc->debugPreStart();
			sqlrc->debugPrint("Caching data to ");
			sqlrc->debugPrint(cachedestname);
			sqlrc->debugPrint("\n");
			sqlrc->debugPreEnd();
		}
	} else {
		if (sqlrc->debug) {
			sqlrc->debugPreStart();
			sqlrc->debugPrint("Resuming caching data to ");
			sqlrc->debugPrint(cachedestname);
			sqlrc->debugPrint("\n");
			sqlrc->debugPreEnd();
		}
	}

	// create the cache file, truncate it unless we're 
	// resuming a previous session
	if (!resumed) {
		cachedestfd=new ofstream(cachedestname,
				ios::in | ios::out | ios::trunc);
		cachedestindfd=new ofstream(cachedestindname,
				ios::in | ios::out | ios::trunc);
	} else {
		cachedestfd=new ofstream(cachedestname,
				ios::in | ios::out | ios::ate | ios::app);
		cachedestindfd=new ofstream(cachedestindname,
				ios::in | ios::out | ios::ate | ios::app);
	}

	if (cachedestfd && cachedestfd->good() && 
				cachedestindfd && cachedestindfd->good()) {

		if (!resumed) {

			// write "magic" identifier to head of files
			cachedestfd->write("SQLRELAYCACHE",13);
			cachedestindfd->write("SQLRELAYCACHE",13);
			
			// write ttl to files
			long	expiration=time(NULL)+cachettl;
			cachedestfd->write((char *)&expiration,sizeof(long));
			cachedestindfd->write((char *)&expiration,sizeof(long));
		}

	} else {

		if (sqlrc->debug) {
			sqlrc->debugPreStart();
			sqlrc->debugPrint("Error caching data to ");
			sqlrc->debugPrint(cachedestname);
			sqlrc->debugPrint("\n");
			sqlrc->debugPreEnd();
		}

		// in case of an error, clean up
		clearCacheDest();
	}
}

void	sqlrcursor::finishCaching() {

	if (sqlrc->debug) {
		sqlrc->debugPreStart();
		sqlrc->debugPrint("Finishing caching.\n");
		sqlrc->debugPreEnd();
	}

	// terminate the result set
	unsigned short	term=END_RESULT_SET;
	cachedestfd->write((char *)&term,sizeof(unsigned short));

	// close the cache file and clean up
	clearCacheDest();
}

void	sqlrcursor::sendInputBinds() {

	if (sqlrc->debug) {
		sqlrc->debugPreStart();
		sqlrc->debugPrint("Sending ");
		sqlrc->debugPrint((long)inbindcount);
		sqlrc->debugPrint(" Input Bind Variables:\n");
		sqlrc->debugPreEnd();
	}

	// write the input bind variables/values to the server.
	sqlrc->writeData((unsigned short)inbindcount);
	unsigned short	size;
	int	count=inbindcount;
	for (int i=0; i<count; i++) {

		// don't send anything if the send flag is turned off
		if (!inbindvars[i].send) {
			count++;
			continue;
		}

		// send the variable
		size=(unsigned short)strlen(inbindvars[i].variable);
		sqlrc->writeData((unsigned short)size);
		sqlrc->writeData(inbindvars[i].variable,(int)size);
		if (sqlrc->debug) {
			sqlrc->debugPreStart();
			sqlrc->debugPrint(inbindvars[i].variable);
			sqlrc->debugPrint("(");
			sqlrc->debugPrint((long)size);
		}

		// send the type
		sqlrc->writeData((unsigned short)inbindvars[i].type);
		if (inbindvars[i].type==NULL_BIND) {
			if (sqlrc->debug) {
				sqlrc->debugPrint(":NULL)\n");
				sqlrc->debugPreEnd();
			}
		} else if (inbindvars[i].type==STRING_BIND) {
			// send the value
			sqlrc->writeData(
				(unsigned short)inbindvars[i].valuesize);
			if (inbindvars[i].valuesize>0) {
				sqlrc->writeData(inbindvars[i].
					value.stringval,
					(int)inbindvars[i].valuesize);
			}

			if (sqlrc->debug) {
				sqlrc->debugPrint(":STRING)=");
				sqlrc->debugPrint(inbindvars[i].
							value.stringval);
				sqlrc->debugPrint("(");
				sqlrc->debugPrint((long)inbindvars[i].
							valuesize);
				sqlrc->debugPrint(")");
				sqlrc->debugPrint("\n");
				sqlrc->debugPreEnd();
			}
		} else if (inbindvars[i].type==LONG_BIND) {
			// send the value
			sqlrc->writeData((unsigned long)inbindvars[i].
								value.longval);

			if (sqlrc->debug) {
				sqlrc->debugPrint(":LONG)=");
				sqlrc->debugPrint((long)inbindvars[i].
							value.longval);
				sqlrc->debugPrint("\n");
				sqlrc->debugPreEnd();
			}
		} else if (inbindvars[i].type==DOUBLE_BIND) {
			// send the value
			sqlrc->writeData((double)inbindvars[i].value.
							doubleval.value);
			sqlrc->writeData(
					(unsigned short)inbindvars[i].value.
							doubleval.precision);
			sqlrc->writeData(
					(unsigned short)inbindvars[i].value.
							doubleval.scale);

			if (sqlrc->debug) {
				sqlrc->debugPrint(":DOUBLE)=");
				sqlrc->debugPrint(inbindvars[i].value.
							doubleval.value);
				sqlrc->debugPrint(":");
				sqlrc->debugPrint((long)inbindvars[i].value.
							doubleval.precision);
				sqlrc->debugPrint(",");
				sqlrc->debugPrint((long)inbindvars[i].value.
							doubleval.scale);
				sqlrc->debugPrint("\n");
				sqlrc->debugPreEnd();
			}
		}
	}
}

void	sqlrcursor::sendOutputBinds() {

	if (sqlrc->debug) {
		sqlrc->debugPreStart();
		sqlrc->debugPrint("Sending Output Bind Variables:\n");
		sqlrc->debugPreEnd();
	}

	// write the output bind variables to the server.
	sqlrc->writeData((unsigned short)outbindcount);
	unsigned short	size;
	int	count=outbindcount;
	for (int i=0; i<count; i++) {

		// don't send anything if the send flag is turned off
		if (!outbindvars[i].send) {
			count++;
			continue;
		}

		// send the variable and size that the buffer needs to be
		size=(unsigned short)strlen(outbindvars[i].variable);
		sqlrc->writeData((unsigned short)size);
		sqlrc->writeData(outbindvars[i].variable,(int)size);
		sqlrc->writeData((unsigned short)outbindvars[i].valuesize);

		if (sqlrc->debug) {
			sqlrc->debugPreStart();
			sqlrc->debugPrint(outbindvars[i].variable);
			sqlrc->debugPrint("(");
			sqlrc->debugPrint((long)outbindvars[i].valuesize);
			sqlrc->debugPrint(")\n");
			sqlrc->debugPreEnd();
		}
	}
}

void	sqlrcursor::sendGetColumnInfo() {

	if (sqlrc->debug) {
		sqlrc->debugPreStart();
		sqlrc->debugPrint("Send Column Info: ");
	}

	if (sendcolumninfo==SEND_COLUMN_INFO) {
		if (sqlrc->debug) {
			sqlrc->debugPrint("yes\n");
		}
		sqlrc->writeData((unsigned short)SEND_COLUMN_INFO);
	} else {
		if (sqlrc->debug) {
			sqlrc->debugPrint("no\n");
		}
		sqlrc->writeData((unsigned short)DONT_SEND_COLUMN_INFO);
	}

	if (sqlrc->debug) {
		sqlrc->debugPreEnd();
	}
}

int	sqlrcursor::getShort(unsigned short *integer) {

	// if the result set is coming from a cache file, read from
	// the file, if not, read from the server
	if (cachesourcefd>-1 && cachesourceindfd>-1) {
		return read(cachesourcefd,(void *)integer,
					sizeof(unsigned short));
	} else {
		return sqlrc->readData(integer);
	}
}

int	sqlrcursor::getLong(unsigned long *integer) {

	// if the result set is coming from a cache file, read from
	// the file, if not, read from the server
	if (cachesourcefd>-1 && cachesourceindfd>-1) {
		return read(cachesourcefd,(void *)integer,
					sizeof(unsigned long));
	} else {
		return sqlrc->readData(integer);
	}
}

int	sqlrcursor::getString(char *string, int size) {

	// if the result set is coming from a cache file, read from
	// the file, if not, read from the server
	if (cachesourcefd>-1 && cachesourceindfd>-1) {
		return read(cachesourcefd,(void *)string,size);
	} else {
		return sqlrc->readData(string,size);
	}
}

void	sqlrcursor::clearColumns() {

	// clear column name strings
	if (columns && sentcolumninfo==SEND_COLUMN_INFO) {
		char	*colname;
		for (int i=0; i<colcount; i++) {
			colname=getColumn(i)->name;
			if (colname && !(colname>=colstorage && 
						colname<=endcolstorageptr)) {
				delete[] colname;
			}
		}
	}

	// reset the column count
	previouscolcount=colcount;
	colcount=0;

	// delete array pointing to each column name
	if (columnnamearray) {
		delete[] columnnamearray;
		columnnamearray=NULL;
	}

	// delete the array for nonstandard types
	if (nonstandardtypes) {
		for (int i=0; i<nonstandardtypescount; i++) {
			delete[] nonstandardtypes[i];
		}
		delete[] nonstandardtypes;
		nonstandardtypes=NULL;
	}

	// reset the column storage pointer
	currentcolstorageptr=colstorage;
}

void	sqlrcursor::dontGetColumnInfo() {
	sendcolumninfo=DONT_SEND_COLUMN_INFO;
}

void	sqlrcursor::getColumnInfo() {
	sendcolumninfo=SEND_COLUMN_INFO;
}

void	sqlrcursor::setResultSetBufferSize(int rows) {
	rsbuffersize=rows;
	if (sqlrc->debug) {
		sqlrc->debugPreStart();
		sqlrc->debugPrint("Result Set Buffer Size: ");
		sqlrc->debugPrint((long)rows);
		sqlrc->debugPrint("\n");
		sqlrc->debugPreEnd();
	}
}

int	sqlrcursor::getResultSetBufferSize() {
	return	rsbuffersize;
}


void	sqlrcursor::cacheToFile(char *filename) {

	cacheon=1;
	cachettl=600;
	if (copyrefs) {
		if (cachedestname) {
			delete[] cachedestname;
		}
		cachedestname=new char[strlen(filename)+1];
		strcpy(cachedestname,filename);
	} else {
		cachedestname=filename;
	}

	// create the index name
	if (cachedestindname) {
		delete[] cachedestindname;
	}
	cachedestindname=new char[strlen(filename)+5];
	sprintf(cachedestindname,"%s.ind",filename);
}

void	sqlrcursor::setCacheTtl(int ttl) {
	cachettl=ttl;
}

void	sqlrcursor::cacheOff() {
	cacheon=0;
}

int	sqlrcursor::getResultSetId() {
	return cursorid;
}

int	sqlrcursor::resumeResultSet(int id) {
	return resumeCachedResultSet(id,NULL);
}

int	sqlrcursor::resumeCachedResultSet(int id, char *filename) {

	if (!endofresultset && !suspendresultsetsent) {
		abortResultSet();
	}
	clearResultSet();

	if (!sqlrc->connected) {
		return 0;
	}

	cached=0;
	resumed=1;
	endofresultset=0;

	if (sqlrc->debug) {
		sqlrc->debugPreStart();
		sqlrc->debugPrint("Resuming Result Set of Cursor: ");
		sqlrc->debugPrint((long)id);
		sqlrc->debugPrint("\n");
		sqlrc->debugPreEnd();
	}

	// tell the server we want to resume the result set
	sqlrc->writeData((unsigned short)RESUME_RESULT_SET);

	// send the id of the cursor we want to 
	// resume the result set of to the server
	sqlrc->writeData((unsigned short)id);

	// process the result set
	if (filename && filename[0]) {
		cacheToFile(filename);
	}
	if (rsbuffersize) {
		if (processResultSet(firstrowindex+rsbuffersize-1)) {
			return 1;
		}
	} else {
		if (processResultSet(-1)) {
			return 1;
		}
	}
	return 0;
}

int	sqlrcursor::sendQuery(char *query) {
	prepareQuery(query);
	return executeQuery();
}

int	sqlrcursor::sendFileQuery(char *path, char *filename) {
	return prepareFileQuery(path,filename) && executeQuery();
}

void	sqlrcursor::initQueryBuffer() {
	if (!querybuffer) {
		querybuffer=new char[MAXQUERYSIZE+1];
		queryptr=querybuffer;
	}
}

void	sqlrcursor::prepareQuery(char *query) {
	reexecute=0;
	validatebinds=0;
	resumed=0;
	clearVariables();
	if (copyrefs) {
		initQueryBuffer();
		strcpy(queryptr,query);
	} else {
		queryptr=query;
	}
}

int	sqlrcursor::prepareFileQuery(char *path, char *filename) {

	// init some variables
	reexecute=0;
	validatebinds=0;
	resumed=0;
	clearVariables();

	// init the fullpath buffer
	if (!fullpath) {
		fullpath=new char[MAXPATHLEN+1];
	}

	// add the path to the fullpath
	int	index=0;
	int	counter=0;
	while (path[index] && counter<MAXPATHLEN) {
		fullpath[counter]=path[index];
		index++;
		counter++;
	}

	// add the "/" to the fullpath
	if (counter<=MAXPATHLEN) {
		fullpath[counter]='/';
		counter++;
	}

	// add the file to the fullpath
	index=0;
	while (filename[index] && counter<MAXPATHLEN) {
		fullpath[counter]=filename[index];
		index++;
		counter++;
	}

	// handle a filename that's too long
	if (counter>MAXPATHLEN) {

		// sabotage the file name so it can't be opened
		fullpath[0]=(char)NULL;

		// debug info
		if (sqlrc->debug) {
			sqlrc->debugPreStart();
			sqlrc->debugPrint("File name ");
			sqlrc->debugPrint(path);
			sqlrc->debugPrint("/");
			sqlrc->debugPrint(filename);
			sqlrc->debugPrint(" is too long.");
			sqlrc->debugPrint("\n");
			sqlrc->debugPreEnd();
		}

	} else {

		// terminate the string
		fullpath[counter]=(char)NULL;

		// debug info
		if (sqlrc->debug) {
			sqlrc->debugPreStart();
			sqlrc->debugPrint("File: ");
			sqlrc->debugPrint(fullpath);
			sqlrc->debugPrint("\n");
			sqlrc->debugPreEnd();
		}
	}

	// open the file
	ifstream queryfile(fullpath);

	// handle a file we couldn't open
	if (!queryfile.good()) {

		// set the error
		char	*err=new char[32+strlen(fullpath)];
		strcpy(err,"The file ");
		strcat(err,fullpath);
		strcat(err," could not be opened.\n");
		if (sqlrc->debug) {
			sqlrc->debugPreStart();
			sqlrc->debugPrint(err);
			sqlrc->debugPreEnd();
		}
		setError(err);
		delete[] err;

		// set queryptr to NULL so executeQuery won't try to do
		// anything with it in the event that it gets called
		queryptr=NULL;

		return 0;
	}

	initQueryBuffer();

	// read the file into the query buffer
	index=0;
	for (;;) {
		if (index==MAXQUERYSIZE) {
			sqlrc->debugPreStart();
			sqlrc->debugPrint("The query in ");
			sqlrc->debugPrint(fullpath);
			sqlrc->debugPrint(" is too large.");
			sqlrc->debugPrint("\n");
			sqlrc->debugPreEnd();
			break;
		}
		querybuffer[index]=queryfile.rdbuf()->sgetc();
		if (querybuffer[index]<=0) {
			break;
		}
		queryfile.rdbuf()->snextc();
		index++;
	}
	querybuffer[index]=0;

	queryfile.close();

	return 1;
}

void	sqlrcursor::substitution(char *variable, char *value) {
	if (subcount<MAXVAR && variable && variable[0]) {
		stringVar(&subvars[subcount],variable,value);
		subcount++;
	}
}

void	sqlrcursor::substitution(char *variable, long value) {
	if (subcount<MAXVAR && variable && variable[0]) {
		longVar(&subvars[subcount],variable,value);
		subcount++;
	}
}

void	sqlrcursor::substitution(char *variable, double value, 
			unsigned short precision, unsigned short scale) {
	if (subcount<MAXVAR && variable && variable[0]) {
		doubleVar(&subvars[subcount],variable,value,precision,scale);
		subcount++;
	}
}

void	sqlrcursor::clearBinds() {
	inbindcount=0;
	outbindcount=0;
}

void	sqlrcursor::inputBind(char *variable, char *value) {
	if (inbindcount<MAXVAR && variable && variable[0]) {
		stringVar(&inbindvars[inbindcount],variable,value);
		inbindvars[inbindcount].send=1;
		inbindcount++;
	}
}

void	sqlrcursor::inputBind(char *variable, unsigned long value) {
	if (inbindcount<MAXVAR && variable && variable[0]) {
		longVar(&inbindvars[inbindcount],variable,value);
		inbindvars[inbindcount].send=1;
		inbindcount++;
	}
}

void	sqlrcursor::inputBind(char *variable, double value, 
				unsigned short precision, 
				unsigned short scale) {
	if (inbindcount<MAXVAR && variable && variable[0]) {
		doubleVar(&inbindvars[inbindcount],variable,value,
						precision, scale);
		inbindvars[inbindcount].send=1;
		inbindcount++;
	}
}

void	sqlrcursor::substitutions(char **variables, char **values) {
	int	index=0;
	while (variables[index] && subcount<MAXVAR) {
		if (variables[index] && variables[index][0]) {
			stringVar(&subvars[subcount],
					variables[index],values[index]);
			subcount++;
		}
		index++;
	}
}

void	sqlrcursor::substitutions(char **variables, long *values) {
	int	index=0;
	while (variables[index] && subcount<MAXVAR) {
		if (variables[index] && variables[index][0]) {
			longVar(&subvars[subcount],
					variables[index],values[index]);
			subcount++;
		}
		index++;
	}
}

void	sqlrcursor::substitutions(char **variables, double *values, 
			unsigned short *precisions, unsigned short *scales) {
	int	index=0;
	while (variables[index] && subcount<MAXVAR) {
		if (variables[index] && variables[index][0]) {
			doubleVar(&subvars[subcount],
					variables[index],
					values[index],
					precisions[index],
					scales[index]);
			subcount++;
		}
		index++;
	}
}

void	sqlrcursor::inputBinds(char **variables, char **values) {
	int	index=0;
	while (variables[index] && inbindcount<MAXVAR) {
		if (variables[index] && variables[index][0]) {
			stringVar(&inbindvars[inbindcount],
					variables[index],values[index]);
			inbindvars[inbindcount].send=1;
			inbindcount++;
		}
		index++;
	}
}

void	sqlrcursor::inputBinds(char **variables, unsigned long *values) {
	int	index=0;
	while (variables[index] && inbindcount<MAXVAR) {
		if (variables[index] && variables[index][0]) {
			longVar(&inbindvars[inbindcount],
					variables[index],values[index]);
			inbindvars[inbindcount].send=1;
			inbindcount++;
		}
		index++;
	}
}

void	sqlrcursor::inputBinds(char **variables, double *values, 
						unsigned short *precisions,
						unsigned short *scales) {
	int	index=0;
	while (variables[index] && inbindcount<MAXVAR) {
		if (variables[index] && variables[index][0]) {
			doubleVar(&inbindvars[inbindcount],
					variables[index],
					values[index],
					precisions[index],
					scales[index]);
			inbindvars[inbindcount].send=1;
			inbindcount++;
		}
		index++;
	}
}

void	sqlrcursor::stringVar(bindvar *var, char *variable, char *value) {

	initVar(var,variable);

	// store the value, handle NULL values too
	if (value) {
		if (copyrefs) {
			var->value.stringval=new char[strlen(value)+1];
			strcpy(var->value.stringval,value);
		} else {
			var->value.stringval=value;
		}
		var->valuesize=strlen(value);
		var->type=STRING_BIND;
	} else {
		var->type=NULL_BIND;
	}
}

void	sqlrcursor::longVar(bindvar *var, char *variable, long value) {
	initVar(var,variable);
	var->type=LONG_BIND;
	var->value.longval=value;
}

void	sqlrcursor::doubleVar(bindvar *var, char *variable, double value,
					unsigned short precision, 
					unsigned short scale) {
	initVar(var,variable);
	var->type=DOUBLE_BIND;
	var->value.doubleval.value=value;
	var->value.doubleval.precision=precision;
	var->value.doubleval.scale=scale;
}

void	sqlrcursor::initVar(bindvar *var, char *variable) {

	// clear any old variable name that was stored and assign the new 
	// variable name also clear any old value that was stored in this 
	// variable
	if (copyrefs) {
		if (var->variable) {
			delete[] var->variable;
		}
		var->variable=new char[strlen(variable)+1];
		strcpy(var->variable,variable);

		if (var->type==STRING_BIND &&
			var->value.stringval) {
			delete[] var->value.stringval;
		}
	} else {
		var->variable=variable;
	}
}

void	sqlrcursor::defineOutputBind(char *variable, int length) {
	if (outbindcount<MAXVAR && variable && variable[0]) {
		if (copyrefs) {
			if (outbindvars[outbindcount].variable) {
				delete[] outbindvars[outbindcount].variable;
			}
			outbindvars[outbindcount].variable=new 
						char[strlen(variable)+1];
			strcpy(outbindvars[outbindcount].variable,variable);
			if (outbindvars[outbindcount].type==STRING_BIND &&
				outbindvars[outbindcount].value.stringval) {
				delete[] outbindvars[outbindcount].
							value.stringval;
			}
		} else {
			outbindvars[outbindcount].variable=variable;
		}
		outbindvars[outbindcount].value.stringval=NULL;
		outbindvars[outbindcount].valuesize=length;
		outbindvars[outbindcount].send=1;
		outbindcount++;
	}
}

char	*sqlrcursor::getOutputBind(char *variable) {

	if (variable) {
		for (int i=0; i<outbindcount; i++) {
			if (!strcmp(outbindvars[i].variable,variable)) {
				return outbindvars[i].value.stringval;
			}
		}
	}
	return NULL;
}

void	sqlrcursor::validateBinds() {
	validatebinds=1;
}

int	sqlrcursor::executeQuery() {

	if (!queryptr) {
		setError("No query to execute.");
		return 0;
	}

	// a useful variable
	int	retval=0;

	if (!subcount) {

		// validate the bind variables
		if (validatebinds) {
			validateBindsInternal(queryptr);
		}
		
		// run the query
		retval=runQuery(queryptr);

	} else {

		// perform substitutions
		strstream	*container=new strstream();
		char		*ptr=queryptr;
		int		found=0;
		int		inquotes=0;
		int		inbraces=0;
		int		len=0;
		strstream	*braces;

		// iterate through the string
		while (*ptr) {
		
			// figure out whether we're inside a quoted 
			// string or not
			if (*ptr=='\'' && *(ptr-1)!='\\') {
				if (inquotes) {
					inquotes=0;
				} else {
					inquotes=1;
				}
			}
		
			// if we find an open-brace then start 
			// sending to a new buffer
			if (*ptr=='[' && !inbraces && !inquotes) {
				braces=new strstream();
				inbraces=1;
				ptr++;
			}
		
			// if we find a close-brace then process 
			// the brace buffer
			if (*ptr==']' && inbraces && !inquotes) {
				*braces << ENDS;
		
				// look for an = sign, skipping whitespace
				char	*bptr=braces->str();
				while (*bptr && (*bptr==' ' || 
					*bptr=='	' || *bptr=='\n')) {
					bptr++;
				}
		
				if (*bptr=='=') {
					// if we find an equals sign first 
					// then process the rest of the buffer
					bptr++;
		
					// skip whitespace
					while (*bptr && (*bptr==' ' || 
						*bptr=='	' || 
					 	*bptr=='\n')) {
						bptr++;
					}
		
					// if the remaining contents of the 
					// buffer are '' or nothing then we 
					// must have an ='' or just an = with 
					// some whitespace, replace this
					// with "is NULL" otherwise, just write
					// out the contents of the buffer
					if (!bptr || 
						(bptr && !strcmp(bptr,"''"))) {
						*container << " is NULL ";
					} else {
						*container << braces->str();
					}
				} else {
					// if we don't find an equals sign, 
					// then write the contents out directly
					*container << braces->str();
				}
				delete braces;
				inbraces=0;
				ptr++;
			}
		
			// if we encounter $(....) then replace the 
			// variable within
			if ((*ptr)=='$' && (*(ptr+1))=='(') {
		
				// first iterate through the arrays passed in
				found=0;
				for (int i=0; i<subcount && !found; i++) {
		
					// if we find a match, write the 
					// value to the container and skip 
					// past the $(variable)
					len=strlen(subvars[i].variable);
					if (!strncmp((ptr+2),
						subvars[i].variable,len) &&
						(*(ptr+2+len))==')') {
		
						if (inbraces) {
							performSubstitution(
								braces,i);
						} else {
							performSubstitution(
								container,i);
						}
						ptr=ptr+3+len;
						found=1;
					}
				}
		
				// if the variable wasn't found, then 
				// just write the $(
				if (!found) {
					if (inbraces) {
						*braces << "$(";
					} else {
						*container << "$(";
					}
					ptr=ptr+2;
				}
		
			} else {
		
				// print out the current character and proceed
				if (inbraces) {
					*braces << *ptr;
				} else {
					*container << *ptr;
				}
				ptr++;
			}
		}
		*container << ENDS;

		// validate the bind variables
		if (validatebinds) {
			validateBindsInternal(container->str());
		}

		// run the query
		retval=runQuery(container->str());

		// clean up
		delete container;
	}

	// set up to re-execute the same query if executeQuery is called
	// again before calling prepareQuery
	reexecute=1;

	return retval;
}

void	sqlrcursor::performSubstitution(strstream *buffer, int which) {

	if (subvars[which].type==STRING_BIND) {
		*buffer << subvars[which].value.stringval;
	} else if (subvars[which].type==LONG_BIND) {
		*buffer << subvars[which].value.longval;
	} else if (subvars[which].type==DOUBLE_BIND) {
		*buffer << setprecision(
				subvars[which].value.doubleval.precision);
		*buffer << floor(subvars[which].value.doubleval.value*
				pow(10,subvars[which].value.doubleval.scale))/
				pow(10,subvars[which].value.doubleval.scale);
	}
}

int	sqlrcursor::runQuery(char *query) {

	// send the query
	if (sendQueryInternal(query)) {

		sendInputBinds();
		sendOutputBinds();
		sendGetColumnInfo();

		if (processResultSet(rsbuffersize-1)) {
			return 1;
		}
	}
	return 0;
}

char	*sqlrcursor::getCacheFileName() {
	return cachedestname;
}

int	sqlrcursor::openCachedResultSet(char *filename) {

	if (sqlrc->debug) {
		sqlrc->debugPreStart();
		sqlrc->debugPrint("Opening cached result set: ");
		sqlrc->debugPrint(filename);
		sqlrc->debugPrint("\n");
		sqlrc->debugPreEnd();
	}

	if (!endofresultset) {
		abortResultSet();
	}
	clearResultSet();

	cached=1;
	endofresultset=0;

	// create the index file name
	char	*indexfilename=new char[strlen(filename)+5];
	sprintf(indexfilename,"%s.ind",filename);

	// open the file
	if ((cachesourcefd=open(filename,O_RDWR|O_EXCL))>-1 &&
		(cachesourceindfd=open(indexfilename,O_RDWR|O_EXCL))>-1) {

		// don't need this anymore
		delete[] indexfilename;

		// initialize firstrowindex and rowcount
		firstrowindex=0;
		rowcount=firstrowindex;

		// make sure it's a cache file and skip the ttl
		char		magicid[13];
		unsigned long	longvar;
		if (getString(magicid,13)==13 &&
			!strncmp(magicid,"SQLRELAYCACHE",13) &&
			getLong(&longvar)==sizeof(unsigned long)) {

			// process the result set
			return processResultSet(firstrowindex+rsbuffersize-1);
		} else {

			// if the test above failed, the file is either not
			// a cache file or is corrupt
			strstream	*errstr=new strstream();
			*errstr << "File " << filename;
			*errstr << " is either corrupt or not a cache file.";
			*errstr << ENDS;
			setError(errstr->str());
			delete errstr;
		}

	} else {

		// if we couldn't open the file, set the error message
		strstream	*errstr=new strstream();
		*errstr << "Couldn't open " << filename;
		*errstr << " and " << indexfilename << ENDS;
		setError(errstr->str());
		delete errstr;
	}

	// don't need this anymore
	delete[] indexfilename;

	// if we fell through to here, then an error has ocurred
	clearCacheSource();
	return 0;
}

int	sqlrcursor::sendQueryInternal(char *query) {

	// if the first 8 characters of the query are "-- debug" followed
	// by a return, then set debugging on
	if (!strncmp(query,"-- debug\n",9)) {
		sqlrc->debugOn();
	}

	if (!endofresultset) {
		abortResultSet();
	}
	clearResultSet();

	if (!sqlrc->connected && !sqlrc->openSession()) {
		return 0;
	}

	cached=0;
	endofresultset=0;

	if (sqlrc->debug) {
		sqlrc->debugPreStart();
		sqlrc->debugPrint("Sending Query:");
		sqlrc->debugPrint("\n");
		sqlrc->debugPrint(query);
		sqlrc->debugPrint("\n");
		sqlrc->debugPrint("Requesting Cursor: ");
		sqlrc->debugPrint((long)cursorid);
		sqlrc->debugPrint("\n");
		sqlrc->debugPreEnd();
	}

	// tell the server we're sending a query
	sqlrc->writeData((unsigned short)SENDING_QUERY);

	// send the cursor id to the server
	sqlrc->writeData((unsigned short)cursorid);

	// send the query to the server.
	if (!reexecute) {
		sqlrc->writeData((unsigned short)DONT_RE_EXECUTE);
		int	querylen=strlen(query);
		sqlrc->writeData((unsigned long)querylen);
		sqlrc->writeData(query,querylen);
	} else {
		sqlrc->writeData((unsigned short)RE_EXECUTE);
	}

	return 1;
}

void	sqlrcursor::validateBindsInternal(char *query) {

	// some useful variables
	char	*ptr;
	char	*start;
	char	*after;
	int	found;
	int	len;
	int	count;

	// check each input bind
	count=inbindcount;
	for (int i=0; i<count; i++) {

		// don't check bind-by-position variables
		len=strlen(inbindvars[i].variable);
		if (isNumber(inbindvars[i].variable,len)) {
			continue;
		}

		found=0;
		start=query+1;

		// there may be more than 1 match for the variable name as in
		// "select * from table where table_name=:table_name", both
		// table_name's would match, but only the second is a bind
		// variable
		while (ptr=strstr(start,inbindvars[i].variable)) {

			// for a match to be a bind variable, it must be 
			// preceded by a colon and can't be followed by an
			// alphabet character, number or underscore
			after=ptr+len;
			if ((*(ptr-1)==':' || *(ptr-1)=='@') && *after!='_' &&
				!(*(after)>='a' && *(after)<='z') &&
				!(*(after)>='A' && *(after)<='Z') &&
				!(*(after)>='0' && *(after)<='9')) {
				found=1;
				break;
			} else {
				// jump past this instance to look for the
				// next one
				start=ptr+len;
			}
		}

		if (!found) {
			inbindvars[i].send=0;
			inbindcount--;
		}
	}

	// check each output bind
	count=outbindcount;
	for (int i=0; i<count; i++) {

		// don't check bind-by-position variables
		len=strlen(outbindvars[i].variable);
		if (isNumber(outbindvars[i].variable,len)) {
			continue;
		}

		found=0;
		start=query+1;

		// there may be more than 1 match for the variable name as in
		// "select * from table where table_name=:table_name", both
		// table_name's would match, but only 1 is correct
		while (ptr=strstr(start,outbindvars[i].variable)) {

			// for a match to be a bind variable, it must be 
			// preceded by a colon and can't be followed by an
			// alphabet character, number or underscore
			after=ptr+len;
			if (*(ptr-1)==':' && *after!='_' &&
				!(*(after)>='a' && *(after)<='z') &&
				!(*(after)>='A' && *(after)<='Z') &&
				!(*(after)>='0' && *(after)<='9')) {
				found=1;
				break;
			} else {
				// jump past this instance to look for the
				// next one
				start=ptr+len;
			}
		}

		if (!found) {
			outbindvars[i].send=0;
			outbindcount--;
		}
	}
}

void	sqlrcursor::clearResultSet() {

	clearCacheDest();
	clearCacheSource();
	clearError();

	// columns is cleared after rows because colcount is used in 
	// clearRows() and set to 0 in clearColumns()
	clearRows();
	clearColumns();

	// clear row counters, since fetchRowIntoBuffer() and clearResultSet()
	// are the only methods that call clearRows() and fetchRowIntoBuffer()
	// needs these values not to be cleared, we'll clear them here...
	firstrowindex=0;
	previousrowcount=rowcount;
	rowcount=0;
	actualrows=0;
	affectedrows=0;
	endofresultset=1;
	suspendresultsetsent=0;
	getrowcount=0;
}

char	*sqlrcursor::getFieldInternal(int row, int col) {
	if (row<OPTIMISTIC_ROW_COUNT) {
		return rows[row]->getField(col);
	}
	return extrarows[row-OPTIMISTIC_ROW_COUNT]->getField(col);
}

unsigned long	sqlrcursor::getFieldLengthInternal(int row, int col) {
	if (row<OPTIMISTIC_ROW_COUNT) {
		return rows[row]->getFieldLength(col);
	}
	return extrarows[row-OPTIMISTIC_ROW_COUNT]->getFieldLength(col);
}

char	*sqlrcursor::getRowStorage(int length) {

	// if there's room in the row storage buffer, allocate it out, 
	// if not, create a new buffer and pass it back
	if (length>endrowstorageptr-currentrowstorageptr) {
		return new char[length];
	}
	char	*retval=currentrowstorageptr;
	currentrowstorageptr=currentrowstorageptr+length;
	return retval;
}

void	sqlrcursor::createRowBuffers() {

	// rows and rowstorage will hang around from now until the cursor is 
	// deleted, getting reused with each query
	rows=new row *[OPTIMISTIC_ROW_COUNT];
	for (int i=0; i<OPTIMISTIC_ROW_COUNT; i++) {
		rows[i]=new row(colcount);
	}

	// Storage is a big buffer that the result set goes into, 
	// optimistically all of it.
	rowstorage=new char[OPTIMISTIC_RESULT_SET_SIZE];
	currentrowstorageptr=rowstorage;
	endrowstorageptr=rowstorage+OPTIMISTIC_RESULT_SET_SIZE;
}

char	*sqlrcursor::getColStorage(int length) {

	// if there's room in the row storage buffer, allocate it out, 
	// if not, create a new buffer and pass it back
	if (length>endcolstorageptr-currentcolstorageptr) {
		return new char[length];
	}
	char	*retval=currentcolstorageptr;
	currentcolstorageptr=currentcolstorageptr+length;


	return retval;
}
