// Copyright (c) 1999-2001  David Muse
// See the file COPYING for more information

#include <config.h>
#include <defaults.h>
#include <defines.h>

#include <configfile.h>
#include <rudiments/commandline.h>
#include <listener.h>

#include <rudiments/permissions.h>

#include <iostream.h>
#include <fstream.h>

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <sys/ipc.h>
#include <stdlib.h>
#include <string.h>
#ifdef HAVE_UNISTD_H
	#include <unistd.h>
#endif
#include <errno.h>

listener::listener() : daemonprocess(), serversocketpool() {

	tmpdirlen=strlen(TMP_DIR);

	initialized=0;

	semset=NULL;
	idmemory=NULL;

	pidfile=NULL;

	id=NULL;
	config=NULL;
	authc=NULL;

	inetport=0;
	unixport=NULL;

	clientsockin=NULL;
	clientsockun=NULL;
	handoffsockun=NULL;
	removehandoffsockun=NULL;

	handoffsocklist=NULL;

	denied=NULL;
	allowed=NULL;

	#ifdef SERVER_DEBUG
	debug=0;
	debugfile=NULL;
	#endif
}

listener::~listener() {
	if (pidfile) {
		unlink(pidfile);
		delete[] pidfile;
	}
	if (initialized) {
		cleanUp();
	}
}

int	listener::initListener(int argc, char **argv) {

	initialized=1;

	// read the commandline
	commandline	*cmdl=new commandline(argc,argv);


	// get the id
	char	*tmpid=cmdl->value("-id");
	if (!(tmpid && tmpid[0])) {
		tmpid=DEFAULT_ID;
		cerr << "Warning! using default id." << ENDL;
	}
	id=new char[strlen(tmpid)+1];
	strcpy(id,tmpid);


	// check/set pid file
	pidfile=new char[tmpdirlen+15+strlen(id)+1];
	sprintf(pidfile,"%s/sqlr-listener-%s",TMP_DIR,id);
	if (checkForPidFile(pidfile)) {
		cerr << ENDL << "sqlr-listener error:" << ENDL;
		cerr << "	The pid file " << TMP_DIR;
		cerr << "/sqlr-listener-" << id << " exists." << ENDL;
		cerr << "	This usually means that the sqlr-listener ";
		cerr << "is already running for the " << ENDL;
		cerr << "	" << id << " instance." << ENDL;
		cerr << "	If it is not running, please remove the file ";
		cerr << "and restart." << ENDL;
		delete[] pidfile;
		pidfile=NULL;
		delete cmdl;
		return 0;
	}
	createPidFile(pidfile);


	// get the config file
	char	*tmpconfig=cmdl->value("-config");
	if (!(tmpconfig && tmpconfig[0])) {
		tmpconfig=DEFAULT_CONFIG_FILE;
	}
	config=new char[strlen(tmpconfig)+1];
	strcpy(config,tmpconfig);


	// parse the config file(s)
	unsigned short	port=0;
	dynamicscaling=0;
	configfile	*cfgfile=new configfile();
	if (cfgfile->parse(config,id)) {

		// run as user/group specified in the config file
		char	*runasuser=cfgfile->getRunAsUser();
		char	*runasgroup=cfgfile->getRunAsGroup();
		if (runasuser[0] && runasgroup[0]) {
			runAsUser(runasuser);
			runAsGroup(runasgroup);
		}

		// get the dynamic connection scaling parameters
		connections=cfgfile->getConnections();
		maxconnections=cfgfile->getMaxConnections();
		maxqueuelength=cfgfile->getMaxQueueLength();
		growby=cfgfile->getGrowBy();
		ttl=cfgfile->getTtl();

		// if dynamic scaling isn't going to be used, disable it
		if (maxqueuelength>=0 && maxconnections>connections &&
			growby>0 && ttl>0) {
			dynamicscaling=1;
		}
		

		// get inet and unix ports to listen on
		port=cfgfile->getPort();
		char	*uport=cfgfile->getUnixPort();
		if (uport && uport[0]) {
			unixport=new char[strlen(uport)+1];
			strcpy(unixport,uport);
		}

		// Get where authentication will be performed. If the 
		// authentication attribute is not "connection" or "database" 
		// then authenticate here.
		char	*authtier=cfgfile->getAuthTier();
		authenticate=0;
		if (strcmp(authtier,"connection") && 
				strcmp(authtier,"database")) {
			authenticate=1;
		}

		// get handoff method
		char	*handoffmethod=cfgfile->getHandOff();
		passdescriptor=0;
		if (!strcmp(handoffmethod,"pass")) {
			passdescriptor=1;

			// create the list of handoff nodes
			handoffsocklist=new handoffsocketnode *[maxconnections];
			for (int i=0; i<maxconnections; i++) {
				handoffsocklist[i]=new handoffsocketnode;
				handoffsocklist[i]->pid=0;
			}
		}

		// get users (if necessary)
		if (authenticate) {
			authc=new authenticator(cfgfile);
		}

		// get denied and allowed ip's and compile the expressions
		char	*deniedips=cfgfile->getDeniedIps();
		char	*allowedips=cfgfile->getAllowedIps();
		if (deniedips[0]) {
			denied=new regularexpression(deniedips);
		}
		if (allowedips[0]) {
			allowed=new regularexpression(allowedips);
		}

		#ifdef SERVER_DEBUG
		openDebugFile();
		#endif
	}


	// handle defaults...
	// if neither port nor unixport are specified, set up to 
	// listen on the default inet port
	if (!port && !unixport) {
		port=DEFAULT_PORT;
		cerr << "Warning! using default port." << ENDL;
	}

	// listen on inet and unix ports
	int	listening=1;
	if (port) {
		clientsockin=new serversocket();
		clientsockin->reuseAddresses();
		listening=clientsockin->listenOnInetPort(port,15);
		if (listening) {
			addSocket(clientsockin);
		} else {
			cerr << "Could not listen on inet port: ";
			cerr << port << ENDL;
			cerr << "Make sure that no other processes are ";
			cerr << "listening on that port.";
			cerr << ENDL << ENDL;
		}
	}

	if (listening && unixport) {
		clientsockun=new serversocket();
		clientsockun->reuseAddresses();
		listening=clientsockun->listenOnUnixPort(unixport,
				permissions::everyoneReadWrite(),15);
		if (listening) {
			addSocket(clientsockun);
		} else {
			cerr << "Could not listen on unix socket: ";
			cerr << unixport << ENDL;
			cerr << "Make sure that the file and ";
			cerr << "directory are readable and writable.";
			cerr << ENDL << ENDL;
		}
	}

	// set the return value of the function
	int	retval=listening;

	if (listening) {

		// If we're going to be passing descriptors to handoff clients
		// to connections, listen on another unix port.  The 
		// connections will connect to this port so the descriptors
		// can be passed.
		if (passdescriptor) {

			// the handoff socket
			char	*handoffsockname=
					new char[tmpdirlen+1+strlen(id)+8+1];
			sprintf(handoffsockname,"%s/%s-handoff",TMP_DIR,id);
			handoffsockun=new serversocket();
			int	success=handoffsockun->listenOnUnixPort(
					handoffsockname,
					permissions::ownerReadWrite(),15);
			if (success) {
				addSocket(handoffsockun);
			} else {
				cerr << "Could not listen on unix socket: ";
				cerr << handoffsockname << ENDL;
				cerr << "Make sure that the file and ";
				cerr << "directory are readable and writable.";
				cerr << ENDL << ENDL;
			}
			delete[] handoffsockname;

			// the deregistration socket
			char	*removehandoffsockname=
					new char[tmpdirlen+1+strlen(id)+14+1];
			sprintf(removehandoffsockname,"%s/%s-removehandoff",
								TMP_DIR,id);
			removehandoffsockun=new serversocket();
			success=removehandoffsockun->listenOnUnixPort(
					removehandoffsockname,
					permissions::ownerReadWrite(),15);
			if (success) {
				addSocket(removehandoffsockun);
			} else {
				cerr << "Could not listen on unix socket: ";
				cerr << removehandoffsockname << ENDL;
				cerr << "Make sure that the file and ";
				cerr << "directory are readable and writable.";
				cerr << ENDL << ENDL;
			}
			delete[] removehandoffsockname;
		}

		// initialize the shared memory segment filename
		char	*idfilename=new char[tmpdirlen+1+strlen(id)+1];
		sprintf(idfilename,"%s/%s",TMP_DIR,id);
		int	idfd=open(idfilename,O_CREAT|O_RDWR,
						permissions::ownerReadWrite());
		if (idfd==-1) {
			cerr << "Could not open: " << idfilename << ENDL;
			cerr << "Make sure that the file and directory are ";
			cerr << "readable and writable." << ENDL << ENDL; 
		} else {
			close(idfd);
		}
	
	
		// create the shared memory segment, if it already exists,
		// attempt to remove and re-create it
		key_t	key=ftok(idfilename,0);
		idmemory=new sharedmemory();
		if (!idmemory->create(key,sizeof(unsigned int)+1024+
						(2*sizeof(unsigned short)),
					permissions::ownerReadWrite())) {

			idmemory->attach(key);
			cerr << ENDL << "sqlr-listener error:" << ENDL;
			cerr << "	Unable to create a shared memory ";
			cerr << "segment.  This is usally because an " << ENDL;
			cerr << "	sqlr-listener is already running for ";
			cerr << "the " << id << " instance.";
			cerr << ENDL << ENDL;
			cerr << "	If it is not running, something may ";
			cerr << "have crashed and left an old segment" << ENDL;
			cerr << "	lying around.  Use the ipcs command ";
			cerr << "to inspect existing shared memory " << ENDL;
			cerr << "	segments and the ipcrm command to ";
			cerr << "remove the shared memory segment with ";
			cerr << ENDL << "	id " << idmemory->getId();
			cerr << "." << ENDL << ENDL;
			retval=0;
			delete idmemory;
		} else {

			// create (or connect) to the semaphore set, if it 
			// already exists, attempt to remove and re-create it
			int	vals[10]={1,1,0,0,1,1,0,0,0,1};
			semset=new semaphoreset();
			if (!semset->create(key,
				permissions::ownerReadWrite(),10,vals)) {

				semset->attach(key,10);
				cerr << ENDL << "sqlr-listener error:" << ENDL;
				cerr << "	Unable to create a semaphore ";
				cerr << "set segment.  This is usally because ";
				cerr << "an " << ENDL;
				cerr << "	sqlr-listener is already ";
				cerr << "running for the " << id;
				cerr << " instance." << ENDL << ENDL;
				cerr << "	If it is not running, ";
				cerr << "something may have crashed and left ";
				cerr << "an old segment" << ENDL;
				cerr << "	lying around.  Use the ipcs ";
				cerr << "command to inspect existing ";
				cerr << "semaphore set " << ENDL;
				cerr << "	segments and the ipcrm ";
				cerr << "command to remove the semaphore set ";
				cerr << "segment with " << ENDL;
				cerr << "	id " << semset->getId();
				cerr << "." << ENDL << ENDL;
				retval=0;
				delete idmemory;
				delete semset;
			}
		}

		delete[] idfilename;
	} else {
		retval=0;
	}


	// clean up
	delete cfgfile;
	delete cmdl;

	// detach from the controlling tty
	if (listening) {
		#ifndef SERVER_DEBUG
		detach();
		#endif
	}

	return retval;
}

void	listener::cleanUp() {

	if (authc) {
		delete authc;
	}

	if (semset) {
		delete semset;
	}

	if (idmemory) {
		delete idmemory;
	}

	if (unixport) {
		delete[] unixport;
	}

	if (id) {
		delete[] id;
	}

	if (config) {
		delete[] config;
	}

	if (clientsockun) {
		delete clientsockun;
	}

	if (clientsockin) {
		delete clientsockin;
	}

	if (handoffsockun) {
		delete handoffsockun;
	}

	if (handoffsocklist) {
		for (int i=0; i<maxconnections; i++) {
			if (handoffsocklist[i]->sock) {
				delete handoffsocklist[i]->sock;
			}
			delete handoffsocklist[i];
		}
		delete[] handoffsocklist;
	}

	if (removehandoffsockun) {
		delete removehandoffsockun;
	}

	if (denied) {
		delete denied;
	}

	if (allowed) {
		delete allowed;
	}

	#ifdef SERVER_DEBUG
	closeDebugFile();
	#endif
}

void	listener::listen() {

	blockSignals();

	for(;;) {


		// accept a tcp connection
		#ifdef SERVER_DEBUG
		debugPrint(0,"waiting for client connection...");
		#endif
		serversocket	*whichsock;
		waitForClientConnection(-1,-1,&whichsock,&clientsock);
		#ifdef SERVER_DEBUG
		debugPrint(0,"done waiting for client connection");
		#endif


		// if the client connected to the handoff or derigistration 
		// socket, handle it and loop back
		if (whichsock==handoffsockun) {
			registerHandoff(clientsock);
			continue;
		} else if (whichsock==removehandoffsockun) {
			deRegisterHandoff(clientsock);
			continue;
		}


		// if the ip was denied, disconnect the socket and loop back
		if (denied && deniedIp()) {
			disconnectClient();
			continue;
		}


		// if the client connected to one of the non-handoff
		// sockets, fork a child to handle it
		long	childpid;
		if (!(childpid=fork())) {

			#ifdef SERVER_DEBUG
			if (debug) {
				closeDebugFile();
				openDebugFile();
			}
			#endif

			// handle authentication
			int	authstatus=getAuth();

			// initialize descriptor pass status
			int	passstatus=0;

			// 3 possible outcomes: 1=pass 0=fail -1=bad data
			if (authstatus==1) {

				if (dynamicscaling) {
					incrementSessionCount();
				}
				getAConnection();
				clientsock->writeData((unsigned short)NO_ERROR);

				// if we're passing file descriptors around,
				// tell the client not to reconnect and pass
				// the descriptor to the appropriate database
				// connection daemon, otherwise tell the client
				// to reconnect and which ports to do it on
				if (passdescriptor) {
					clientsock->writeData(
						(unsigned short)DONT_RECONNECT);

					// if pass descriptor fails, we can
					// still write to the client, send it
					// an error message
					if (!(passstatus=passDescriptor())) {
						clientsock->writeData(
							(unsigned short)ERROR);
						clientsock->writeData(
							(unsigned short)70);
						clientsock->writeData("The listener failed to hand the client off to the database connection.");
					}
				} else {
					clientsock->writeData(
						(unsigned short)RECONNECT);
					clientsock->writeData(unixportstrlen);
					clientsock->writeData(unixportstr);
					clientsock->writeData(
						(unsigned short)inetport);
				}

			} else if (authstatus==0) {

				// sleep before and after returning an
				// authentication error to discourage
				// brute-force password attacks
				sleep(2);
				clientsock->writeData((unsigned short)ERROR);
				sleep(2);
			}

			// Sometimes the listener sends the ports and closes
			// the socket while they are still buffered but not
			// yet transmitted.  This causes the client to receive
			// partial data or an error.  Telling the socket to
			// linger doesn't always fix it.  Doing a read here
			// should guarantee that the client will close it's end
			// of the connection before the server closes it's end;
			// the server will wait for data from the client
			// (which it will never receive) and when the client
			// closes it's end (which it will only do after getting
			// the ports), the read will fall through.  This should
			// guarantee that the client will get the ports without
			// requiring the client to send data back indicating so.

			unsigned short	dummy;
			if (!passdescriptor || 
					(passdescriptor && !authstatus)) {

				// If we're not passing descriptors or if
				// we are but authentication failed, the client
				// shouldn't be sending any data, so a single
				// read should suffice.
				clientsock->readData(&dummy);

			} else if (!passstatus) {

				// If the descriptor pass failed, the client
				// cound send an entire query and bind vars
				// before it reads the error and closes the
				// socket.  We have to absorb all of that
				// data.  We shouldn't just loop forever
				// though, that would provide a point of entry
				// for a DOS attack.  We'll read the maximum
				// number of bytes that could be sent.

				int	counter=0;
				while (clientsock->readData(&dummy)>0 &&
					counter< // sending auth
						(sizeof(short)+
						// user/password
						2*(sizeof(long)+USERSIZE)+
						// sending query
						sizeof(short)+
						// need a cursor
						sizeof(short)+
						// executing new query
						sizeof(short)+
						// query size and query
						sizeof(long)+MAXQUERYSIZE+
						// input bind var count
						sizeof(short)+
						// input bind vars
						MAXVAR*(2*sizeof(short)+
							BINDVARLENGTH)+
						// output bind var count
						sizeof(short)+
						// output bind vars
						MAXVAR*(2*sizeof(short)+
							BINDVARLENGTH)+
						// get column info
						sizeof(short)+
						// skip/fetch
						2*sizeof(long)
						// divide by two because we're
						// reading 2 bytes at a time
						)/2) {
					counter++;
				}
			}

			// clean up and exit
			disconnectClient();

			// since this is the forked off listener, we don't
			// want to actually remove the semaphore set or shared
			// memory segment
			idmemory->dontRemove();
			semset->dontRemove();

			cleanUp();
			_exit(0);
		}

		#ifdef SERVER_DEBUG
		debugPrintStart();
		debugTab(0);
		debugPrint("forked a child: ");
		debugPrint(childpid);
		debugPrintEnd();
		#endif

		// the main process doesn't need to stay connected 
		// to the client, only the forked process
		clientsock->closeSocket();
		delete clientsock;
	}
}

void	listener::blockSignals() {

	// the daemon class handles SIGTERM's and SIGINT's and SIGCHLDS
	// so we're going to block all other signals 
	signalset	set;
	set.removeAllSignals();

	#ifdef HAVE_SIGHUP
		set.addSignal(SIGHUP);
	#endif
	#ifdef HAVE_SIGQUIT
		set.addSignal(SIGQUIT);
	#endif
	#ifdef HAVE_SIGILL
		set.addSignal(SIGILL);
	#endif
	#ifdef HAVE_SIGTRAP
		set.addSignal(SIGTRAP);
	#endif
	#ifdef HAVE_SIGABRT
		set.addSignal(SIGABRT);
	#endif
	#ifdef HAVE_SIGIOT
		set.addSignal(SIGIOT);
	#endif
	#ifdef HAVE_SIGBUS
		set.addSignal(SIGBUS);
	#endif
	#ifdef HAVE_SIGFPE
		set.addSignal(SIGFPE);
	#endif
	#ifdef HAVE_SIGUSR1
		set.addSignal(SIGUSR1);
	#endif
	#ifdef HAVE_SIGSEGV
		set.addSignal(SIGSEGV);
	#endif
	#ifdef HAVE_SIGUSR2
		set.addSignal(SIGUSR2);
	#endif
	#ifdef HAVE_SIGPIPE
		set.addSignal(SIGPIPE);
	#endif
	#ifdef HAVE_SIGALRM
		set.addSignal(SIGALRM);
	#endif
	#ifdef HAVE_SIGSTKFLT
		set.addSignal(SIGSTKFLT);
	#endif
	#ifdef HAVE_SIGCONT
		set.addSignal(SIGCONT);
	#endif
	#ifdef HAVE_SIGSTOP
		set.addSignal(SIGSTOP);
	#endif
	#ifdef HAVE_SIGTSTP
		set.addSignal(SIGTSTP);
	#endif
	#ifdef HAVE_SIGTTIN
		set.addSignal(SIGTTIN);
	#endif
	#ifdef HAVE_SIGTTOU
		set.addSignal(SIGTTOU);
	#endif
	#ifdef HAVE_SIGURG
		set.addSignal(SIGURG);
	#endif
	#ifdef HAVE_SIGXCPU
		set.addSignal(SIGXCPU);
	#endif
	#ifdef HAVE_SIGXFSZ
		set.addSignal(SIGXFSZ);
	#endif
	#ifdef HAVE_SIGVTALRM
		set.addSignal(SIGVTALRM);
	#endif
	#ifdef HAVE_SIGPROF
		set.addSignal(SIGPROF);
	#endif
	#ifdef HAVE_SIGWINCH
		set.addSignal(SIGWINCH);
	#endif
	#ifdef HAVE_SIGIO
		set.addSignal(SIGIO);
	#endif
	#ifdef HAVE_SIGPOLL
		set.addSignal(SIGPOLL);
	#endif
	#ifdef HAVE_SIGPWR
		set.addSignal(SIGPWR);
	#endif
	#ifdef HAVE_SIGUNUSED
		set.addSignal(SIGUNUSED);
	#endif
	#ifdef HAVE_SIGEMT
		set.addSignal(SIGEMT);
	#endif
	#ifdef HAVE_SIGSYS
		set.addSignal(SIGSYS);
	#endif
	#ifdef HAVE_SIGWAITING
		set.addSignal(SIGWAITING);
	#endif
	#ifdef HAVE_SIGLWP
		set.addSignal(SIGLWP);
	#endif
	#ifdef HAVE_SIGFREEZE
		set.addSignal(SIGFREEZE);
	#endif
	#ifdef HAVE_SIGTHAW
		set.addSignal(SIGTHAW);
	#endif
	#ifdef HAVE_SIGCANCEL
		set.addSignal(SIGCANCEL);
	#endif
	#ifdef HAVE_SIGLOST
		set.addSignal(SIGLOST);
	#endif
	#ifdef HAVE__SIGRTMIN
		set.addSignal(_SIGRTMIN);
	#endif
	#ifdef HAVE__SIGRTMAX
		set.addSignal(_SIGRTMAX);
	#endif
	#ifdef HAVE_SIGRTMIN
		set.addSignal(SIGRTMIN);
	#endif
	#ifdef HAVE_SIGRTMAX
		set.addSignal(SIGRTMAX);
	#endif

	signalmanager::ignoreSignals(set.getSignalSet());
}

void	listener::registerHandoff(genericsocket *sock) {

	#ifdef SERVER_DEBUG
	debugPrint(0,"registering handoff...");
	#endif

	// get the connection daemon's pid
	unsigned long processid;
	sock->readData(&processid);

	// find a free node in the list, if we find another node with the
	// same pid, then the old connection must have died off mysteriously,
	// replace it
	for (int i=0; i<maxconnections; i++) {
		if (!handoffsocklist[i]->pid) {
			handoffsocklist[i]->pid=processid;
			handoffsocklist[i]->sock=sock;
			break;
		} else if (handoffsocklist[i]->pid==processid) {
			handoffsocklist[i]->sock=sock;
			break;
		}
	}

	#ifdef SERVER_DEBUG
	debugPrint(0,"done registering handoff...");
	#endif
}

void	listener::deRegisterHandoff(genericsocket *sock) {

	#ifdef SERVER_DEBUG
	debugPrint(0,"de-registering handoff...");
	#endif

	// get the connection daemon's pid
	unsigned long	processid;
	sock->readData(&processid);

	// remove the matching socket from the list
	for (int i=0; i<maxconnections; i++) {
		if (handoffsocklist[i]->pid==processid) {
			handoffsocklist[i]->pid=0;
			delete handoffsocklist[i]->sock;
			#ifdef SERVER_DEBUG
			debugPrint(0,"done de-registering handoff...");
			#endif
			return;
		}
	}

	// clean up
	delete clientsock;

	#ifdef SERVER_DEBUG
	debugPrint(0,"done de-registering handoff...");
	#endif
}

int	listener::deniedIp() {

	#ifdef SERVER_DEBUG
	debugPrint(0,"checking for valid ip...");
	#endif

	char	ip[16];
	if (serversocket::getClientAddress(clientsock,ip) &&
			denied->match(ip) && 
			(!allowed || (allowed && !allowed->match(ip)))) {

		#ifdef SERVER_DEBUG
		debugPrint(0,"invalid ip...");
		#endif

		return 1;
	}

	#ifdef SERVER_DEBUG
	debugPrint(0,"valid ip...");
	#endif

	return 0;
}

void	listener::disconnectClient() {
	clientsock->closeSocket();
	delete clientsock;
}

int	listener::getAuth() {

	#ifdef SERVER_DEBUG
	debugPrint(0,"getting authentication...");
	#endif

	// Get the user/password. For either one, if they are too big or
	// if there's a read error, just exit with an error code
	unsigned long	size;
	clientsock->readData(&size);
	if (size>USERSIZE || clientsock->readData(userbuffer,size)!=size) {
		#ifdef SERVER_DEBUG
		debugPrint(0,"authentication failed: user size is wrong");
		#endif
		return -1;
	}
	userbuffer[size]=(char)NULL;
	clientsock->readData(&size);
	if (size>USERSIZE || clientsock->readData(passwordbuffer,size)!=size) {
		#ifdef SERVER_DEBUG
		debugPrint(0,"authentication failed: password size is wrong");
		#endif
		return -1;
	}
	passwordbuffer[size]=(char)NULL;

	// If the listener is supposed to authenticate, then do so, otherwise
	// just return 1 as if authentication succeeded.
	if (authenticate) {

		// Return 1 if what the client sent matches one of the 
		// user/password sets and 0 if no match is found.
		int	retval=authc->authenticate(userbuffer,passwordbuffer);
		#ifdef SERVER_DEBUG
		if (retval) {
			debugPrint(1,"listener-based authentication succeeded");
		} else {
			debugPrint(1,"listener-based authentication failed: invalid user/password");
		}
		#endif
		return retval;
	}

	#ifdef SERVER_DEBUG
	debugPrint(0,"done getting authentication");
	#endif

	return 1;
}

void	listener::getAConnection() {

	#ifdef SERVER_DEBUG
	debugPrint(0,"getting a connection...");
	#endif

	// wait on the read mutex
	semset->wait(1);

	// wait on a writer
	semset->wait(2);

	// get a pointer to the shared memory segment
	char	*ptr=(char *)((long)idmemory->getPointer()+
					(2*sizeof(unsigned int)));

	// if we're passing descriptors around, the connection will pass it's
	// pid to us, otherwise it will pass it's inet and unix ports
	if (passdescriptor) {

		#ifdef SERVER_DEBUG
		debugPrint(1,"handoff=pass");
		#endif

		// get the pid
		pid=*((unsigned long *)ptr);

	} else {

		#ifdef SERVER_DEBUG
		debugPrint(1,"handoff=reconnect");
		#endif

		// buffer to hold the port
		unsigned short	connectionport;

		// get the unix port
		unixportstrlen=0;
		while (*ptr!=':' && unixportstrlen<MAXPATHLEN) {
			unixportstr[unixportstrlen]=*ptr;
			unixportstrlen++;
			ptr++;
		}
		unixportstr[unixportstrlen]=(char)NULL;
		ptr++;

		// get the inet port (it will be right-aligned)
		inetport=*((unsigned char *)ptr) | 
				(*((unsigned char *)(ptr+1)) << 8);

	}

	// done reading
	semset->signal(3);

	// signal on the read mutex
	semset->signal(1);

	#ifdef SERVER_DEBUG
	debugPrint(0,"done getting a connection");
	#endif
	
}

int	listener::passDescriptor() {

	#ifdef SERVER_DEBUG
	debugPrint(0,"passing descriptor...");
	#endif

	// Look through the list of handoff sockets for the pid of the 
	// connection that we got during the call to getAConnection().
	// When we find it, send the descriptor of the clientsock to the 
	// connection over the handoff socket associated with that node.
	//
	// If for some reason, the socket is closed when we find it, remove it
	// and return an error.

	for (int i=0; i<maxconnections; i++) {
		if (handoffsocklist[i]->pid==pid) {
			if (handoffsocklist[i]->sock->passFileDescriptor(
						clientsock->getDescriptor())) {

				#ifdef SERVER_DEBUG
				debugPrint(0,"done passing descriptor");
				#endif
				return 1;
			} 
			handoffsocklist[i]->pid=0;
			delete handoffsocklist[i]->sock;
			break;
		}
	}

	#ifdef SERVER_DEBUG
	debugPrint(0,"passing descriptor failed");
	#endif
	return 0;
}

void	listener::incrementSessionCount() {

	#ifdef SERVER_DEBUG
	debugPrint(0,"inrementing session count...");
	#endif

	// wait for access
	semset->wait(5);

	// increment the counter
	unsigned int	*sessioncount=
				(unsigned int *)((long)idmemory->getPointer()+
					sizeof(unsigned int));
	(*sessioncount)++;

	#ifdef SERVER_DEBUG
	debugPrint(1,(long)(*sessioncount));
	#endif

	if (dynamicscaling) {
		// signal the scaler
		semset->signal(6);

		// wait for the scaler
		semset->wait(7);
	}

	// signal that others may have access
	semset->signal(5);

	#ifdef SERVER_DEBUG
	debugPrint(0,"done inrementing session count");
	#endif
}

char	*listener::getUnixServerPort() {
	return unixport;
}

#ifdef SERVER_DEBUG
void	listener::openDebugFile() {

	char	*debugfilename=new char[strlen(DEBUG_DIR)+15+20+1];
	sprintf(debugfilename,"%s/sqlr-listener.%d",DEBUG_DIR,getpid());

	mode_t	oldumask=umask(066);
	debugfile=new ofstream(debugfilename);
	umask(oldumask);

	if (debugfile && debugfile->good()) {
		cout << "Debugging to: " << debugfilename << ENDL;
		debug=1;
	} else {
		cout << "Couldn't open debug file: ";
		cout << debugfilename << ENDL;
		if (debugfile) {
			debugfile->close();
			delete debugfile;
			debugfile=NULL;
		}
		debug=0;
	}

	delete[] debugfilename;
}

void	listener::closeDebugFile() {
	if (debugfile) {
		debugfile->close();
		delete debugfile;
	}
}

void	listener::debugPrintStart() {
	if (!debug) {
		return;
	}
	*debugfile << "listener " << getpid() << ": ";
}

void	listener::debugTab(int tab) {
	if (!debug) {
		return;
	}
	for (int i=0; i<tab; i++) {
		*debugfile << "	";
	}
}

void	listener::debugPrint(char *string) {
	if (!debug) {
		return;
	}
	*debugfile << string;
}

void	listener::debugPrint(long number) {
	if (!debug) {
		return;
	}
	*debugfile << number;
}

void	listener::debugPrint(double number) {
	if (!debug) {
		return;
	}
	*debugfile << number;
}

void	listener::debugPrint(int tab, char *string) {
	if (!debug) {
		return;
	}
	debugPrintStart();
	debugTab(tab);
	*debugfile << string;
	debugPrintEnd();
}

void	listener::debugPrint(int tab, long number) {
	if (!debug) {
		return;
	}
	debugPrintStart();
	debugTab(tab);
	*debugfile << number;
	debugPrintEnd();
}

void	listener::debugPrint(int tab, double number) {
	if (!debug) {
		return;
	}
	debugPrintStart();
	debugTab(tab);
	*debugfile << number;
	debugPrintEnd();
}

void	listener::debugPrintEnd() {
	if (!debug) {
		return;
	}
	*debugfile << ENDL;
}
#endif
