// Copyright (c) 1999-2001  David Muse
// See the file COPYING for more information

#include <config.h>
#include <defaults.h>
#include <rudiments/commandline.h>
#include <scaler.h>

#include <rudiments/permissions.h>

#include <iostream.h>

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <sys/ipc.h>
#include <stdlib.h>
#include <time.h>
#include <string.h>

#ifdef HAVE_UNISTD_H
	#include <unistd.h>
#endif

scaler::scaler() : daemonprocess() {

	tmpdirlen=strlen(TMP_DIR);
	init=0;

	pidfile=NULL;
	semset=NULL;
	idmemory=NULL;
	cfgfile=NULL;

	id=NULL;
	config=NULL;
	dbase=NULL;
}

scaler::~scaler() {
	if (init) {
		cleanUp();
	}
}

int	scaler::initScaler(int argc, char **argv) {

	init=1;

	// read the commandline
	commandline	*cmdl=new commandline(argc,argv);

	// get the id
	char	*tmpid=cmdl->value("-id");
	if (!(tmpid && tmpid[0])) {
		tmpid=DEFAULT_ID;
		cerr << "Warning! using default id." << ENDL;
	}
	id=new char[strlen(tmpid)+1];
	strcpy(id,tmpid);

	// check for listener's pid file
	char	*listenerpidfile=new char[tmpdirlen+15+strlen(id)+1];
	sprintf(listenerpidfile,"%s/sqlr-listener-%s",TMP_DIR,id);
	if (!checkForPidFile(listenerpidfile)) {
		cerr << ENDL << "sqlr-scaler error: " << ENDL;
		cerr << "	The file " << TMP_DIR;
		cerr << "/sqlr-listener-" << id;
		cerr << " was not found." << ENDL;
		cerr << "	This usually means that the sqlr-listener ";
		cerr << "is not running." << ENDL;
		cerr << "	The sqlr-listener must be running ";
		cerr << "for the sqlr-scaler to start." << ENDL << ENDL;
		delete cmdl;
		delete[] listenerpidfile;
		return 0;
	}
	delete[] listenerpidfile;

	// check/set pid file
	pidfile=new char[tmpdirlen+13+strlen(id)+1];
	sprintf(pidfile,"%s/sqlr-scaler-%s",TMP_DIR,id);
	if (checkForPidFile(pidfile)) {
		cerr << ENDL << "sqlr-scaler error:" << ENDL;
		cerr << "	The pid file " << TMP_DIR;
		cerr << "/sqlr-scaler-" << id;
		cerr << " exists." << ENDL;
		cerr << "	This usually means that the sqlr-scaler ";
		cerr << "is already running for the " << ENDL;
		cerr << "	" << id << " instance." << ENDL;
		cerr << "	If it is not running, please remove the file ";
		cerr << "and restart." << ENDL;
		delete[] pidfile;
		pidfile=NULL;
		delete cmdl;
		return 0;
	}
	createPidFile(pidfile);

	// get the config file
	char	*tmpconfig=cmdl->value("-config");
	if (!(tmpconfig && tmpconfig[0])) {
		tmpconfig=DEFAULT_CONFIG_FILE;
	}
	config=new char[strlen(tmpconfig)+1];
	strcpy(config,tmpconfig);

	// parse the config file
	cfgfile=new configfile();

	
	// parse the system-wide config file, then the user's personal file
	if (cfgfile->parse(config,id)) {

		// run as user/group specified in the config file
		char	*runasuser=cfgfile->getRunAsUser();
		char	*runasgroup=cfgfile->getRunAsGroup();
		if (runasuser[0] && runasgroup[0]) {
			runAsUser(runasuser);
			runAsGroup(runasgroup);
		}

		// get the dynamic connection scaling parameters
		connections=cfgfile->getConnections();
		maxconnections=cfgfile->getMaxConnections();
		maxqueuelength=cfgfile->getMaxQueueLength();
		growby=cfgfile->getGrowBy();
		ttl=cfgfile->getTtl();

		// get the database type
		dbase=new char[strlen(cfgfile->getDbase())+1];
		strcpy(dbase,cfgfile->getDbase());

		// get the list of connect strings
		connectstringlist=cfgfile->getConnectStrings();

		// add up the connection metrics
		connectstringnode	*currentnode=connectstringlist;
		metrictotal=0;
		while (currentnode) {
			metrictotal=metrictotal+currentnode->getMetric();
			currentnode=currentnode->getNext();
		}
	}


	// initialize the shared memory segment filename
	idfilename=new char[tmpdirlen+1+strlen(id)+1];
	sprintf(idfilename,"%s/%s",TMP_DIR,id);
	key_t	key=ftok(idfilename,0);

	// connect to the semaphore set
	semset=new semaphoreset();
	semset->showErrors();
	semset->attach(key,10);

	// connect to the shared memory segment
	idmemory=new sharedmemory();
	idmemory->attach(key);

	// set up random number generator
	currentseed=time(NULL);

	// clean up
	delete cmdl;

	// detach from the controlling tty
	detach();

	return 1;
}

void	scaler::cleanUp() {

	delete[] idfilename;

	if (semset) {
		delete semset;
	}

	if (idmemory) {
		delete idmemory;
	}

	if (id) {
		delete[] id;
	}

	if (pidfile) {
		unlink(pidfile);
		delete[] pidfile;
	}

	if (config) {
		delete[] config;
	}

	if (dbase) {
		delete[] dbase;
	}

	if (cfgfile) {
		delete cfgfile;
	}
}


void	scaler::openMoreConnections() {

	// wait for a new listener to fire up and increment the listener count
	semset->wait(6);

	// maxqueuelength<0, maxconnections<=connections, growby<=0 or ttl<=0
	// would disable this whole feature (or cause problems) so
	// catch them up front
	if (maxqueuelength<0 || maxconnections<=connections || 
		growby<=0 || ttl<=0) {
		semset->signal(7);
		return;
	}

	int	sessions=countSessions();
	int	connections=countConnections();

	// signal listener to keep going
	semset->signal(7);

	// do we need to open more connections?
	if ((sessions-connections)<=maxqueuelength) {
		return;
	}

	// can more be opened, or will we exceed the max?
	if ((connections+growby)<=maxconnections) {

		// open "growby" connections
		for (int i=0; i<growby; i++) {

			getRandomConnectionId();

			// build the command to start a connection
			char	*command=new char[16+strlen(dbase)+
						6+20+
						5+strlen(id)+
						15+strlen(connectionid)+
						9+strlen(config)+1];
			sprintf(command,
		"sqlr-connection-%s -ttl %d -id %s -connectionid %s -config %s",
				dbase,ttl,id,connectionid,config);

			// start another connection
			system(command);

			// clean up
			delete[] command;

			// wait for the connection count to increase
			semset->wait(8);
		}
	}
}

void	scaler::getRandomConnectionId() {

	// get a scaled random number
	currentseed=randomnumber::generateNumber(currentseed);
	int	scalednum=randomnumber::scaleNumber(currentseed,0,metrictotal);

	// run through list, decrementing scalednum by the metric
	// for each, when scalednum is 0, pick that connection id
	connectstringnode	*currentnode=connectstringlist;
	while (currentnode) {
		scalednum=scalednum-currentnode->getMetric();
		if (scalednum<=0) {
			connectionid=currentnode->getConnectionId();
			break;
		}
		currentnode=currentnode->getNext();
	}
}

int	scaler::countSessions() {

	// get the number of open connections
	unsigned int	*sessions=(unsigned int *)((long)idmemory->getPointer()+
							sizeof(unsigned int));

	return (int)(*sessions);
}

int	scaler::countConnections() {

	// wait for access to the connection counter
	semset->wait(4);

	// get the number of connections
	unsigned int	*connections=(unsigned int *)idmemory->getPointer();

	// signal that the connection counter may be accessed by someone else
	semset->signal(4);

	return (int)(*connections);
}

void	scaler::loop() {

	for (;;) {
		// open more connections if necessary
		openMoreConnections();
	}
}
